# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from oslo_utils import timeutils

from nova import context
from nova.notifications.objects import service as service_notification
from nova import objects
from nova.objects import fields
from nova import test
from nova.tests.unit.objects.test_service import fake_service


class TestServiceStatusNotification(test.TestCase):
    def setUp(self):
        self.ctxt = context.get_admin_context()
        super(TestServiceStatusNotification, self).setUp()

    @mock.patch('nova.notifications.objects.service.ServiceStatusNotification')
    def _verify_notification(self, service_obj, action, mock_notification):
        if action == fields.NotificationAction.CREATE:
            service_obj.create()
        elif action == fields.NotificationAction.UPDATE:
            service_obj.save()
        elif action == fields.NotificationAction.DELETE:
            service_obj.destroy()
        else:
            raise Exception('Unsupported action: %s' % action)

        self.assertTrue(mock_notification.called)

        event_type = mock_notification.call_args[1]['event_type']
        priority = mock_notification.call_args[1]['priority']
        publisher = mock_notification.call_args[1]['publisher']
        payload = mock_notification.call_args[1]['payload']

        self.assertEqual(service_obj.host, publisher.host)
        self.assertEqual(service_obj.binary, publisher.source)
        self.assertEqual(fields.NotificationPriority.INFO, priority)
        self.assertEqual('service', event_type.object)
        self.assertEqual(action, event_type.action)
        for field in service_notification.ServiceStatusPayload.SCHEMA:
            if field in fake_service:
                self.assertEqual(fake_service[field], getattr(payload, field))

        mock_notification.return_value.emit.assert_called_once_with(self.ctxt)

    @mock.patch('nova.db.main.api.service_update')
    def test_service_update_with_notification(self, mock_db_service_update):
        service_obj = objects.Service(context=self.ctxt, id=fake_service['id'])
        mock_db_service_update.return_value = fake_service
        for key, value in {'disabled': True,
                           'disabled_reason': 'my reason',
                           'forced_down': True}.items():
            setattr(service_obj, key, value)
            self._verify_notification(service_obj,
                                      fields.NotificationAction.UPDATE)

    @mock.patch('nova.notifications.objects.service.ServiceStatusNotification')
    @mock.patch('nova.db.main.api.service_update')
    def test_service_update_without_notification(self,
                                                 mock_db_service_update,
                                                 mock_notification):
        service_obj = objects.Service(context=self.ctxt, id=fake_service['id'])
        mock_db_service_update.return_value = fake_service

        for key, value in {'report_count': 13,
                           'last_seen_up': timeutils.utcnow()}.items():
            setattr(service_obj, key, value)
            service_obj.save()
            self.assertFalse(mock_notification.called)

    @mock.patch('nova.db.main.api.service_create')
    def test_service_create_with_notification(self, mock_db_service_create):
        service_obj = objects.Service(context=self.ctxt)
        service_obj["uuid"] = fake_service["uuid"]
        mock_db_service_create.return_value = fake_service
        self._verify_notification(service_obj,
                                  fields.NotificationAction.CREATE)

    @mock.patch('nova.db.main.api.service_destroy')
    def test_service_destroy_with_notification(self, mock_db_service_destroy):
        service = copy.deepcopy(fake_service)
        service.pop("version")
        service_obj = objects.Service(context=self.ctxt, **service)
        self._verify_notification(service_obj,
                                  fields.NotificationAction.DELETE)



# Copyright 2011 OpenStack Foundation
# Copyright 2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
import webob

from nova.api.openstack.compute import admin_password as admin_password_v21
from nova import exception
from nova import test
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_instance


def fake_get(self, context, id, expected_attrs=None,
             cell_down_support=False):
    return fake_instance.fake_instance_obj(
        context,
        uuid=id,
        project_id=context.project_id,
        user_id=context.user_id,
        expected_attrs=expected_attrs)


def fake_set_admin_password(self, context, instance, password=None):
    pass


class AdminPasswordTestV21(test.NoDBTestCase):
    validation_error = exception.ValidationError

    def setUp(self):
        super(AdminPasswordTestV21, self).setUp()
        self.stub_out('nova.compute.api.API.set_admin_password',
                      fake_set_admin_password)
        self.stub_out('nova.compute.api.API.get', fake_get)
        self.fake_req = fakes.HTTPRequest.blank('')

    def _get_action(self):
        return admin_password_v21.AdminPasswordController().change_password

    def _check_status(self, expected_status, res, controller_method):
        self.assertEqual(expected_status, controller_method.wsgi_code)

    def test_change_password(self):
        body = {'changePassword': {'adminPass': 'test'}}
        res = self._get_action()(self.fake_req, fakes.FAKE_UUID, body=body)
        self._check_status(202, res, self._get_action())

    def test_change_password_empty_string(self):
        body = {'changePassword': {'adminPass': ''}}
        res = self._get_action()(self.fake_req, fakes.FAKE_UUID, body=body)
        self._check_status(202, res, self._get_action())

    @mock.patch('nova.compute.api.API.set_admin_password',
                side_effect=NotImplementedError())
    def test_change_password_with_non_implement(self, mock_set_admin_password):
        body = {'changePassword': {'adminPass': 'test'}}
        self.assertRaises(webob.exc.HTTPNotImplemented,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    @mock.patch('nova.compute.api.API.get',
                side_effect=exception.InstanceNotFound(
                    instance_id=fakes.FAKE_UUID))
    def test_change_password_with_non_existed_instance(self, mock_get):
        body = {'changePassword': {'adminPass': 'test'}}
        self.assertRaises(webob.exc.HTTPNotFound,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    def test_change_password_with_non_string_password(self):
        body = {'changePassword': {'adminPass': 1234}}
        self.assertRaises(self.validation_error,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    @mock.patch('nova.compute.api.API.set_admin_password',
                side_effect=exception.InstancePasswordSetFailed(instance="1",
                                                                reason=''))
    def test_change_password_failed(self, mock_set_admin_password):
        body = {'changePassword': {'adminPass': 'test'}}
        self.assertRaises(webob.exc.HTTPConflict,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    @mock.patch('nova.compute.api.API.set_admin_password',
                side_effect=exception.SetAdminPasswdNotSupported(instance="1",
                                                                 reason=''))
    def test_change_password_not_supported(self, mock_set_admin_password):
        body = {'changePassword': {'adminPass': 'test'}}
        self.assertRaises(webob.exc.HTTPConflict,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    @mock.patch('nova.compute.api.API.set_admin_password',
                side_effect=exception.InstanceAgentNotEnabled(instance="1",
                                                              reason=''))
    def test_change_password_guest_agent_disabled(self,
                                                  mock_set_admin_password):
        body = {'changePassword': {'adminPass': 'test'}}
        self.assertRaises(webob.exc.HTTPConflict,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    def test_change_password_without_admin_password(self):
        body = {'changPassword': {}}
        self.assertRaises(self.validation_error,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    def test_change_password_none(self):
        body = {'changePassword': {'adminPass': None}}
        ex = self.assertRaises(self.validation_error,
                               self._get_action(),
                               self.fake_req, fakes.FAKE_UUID, body=body)
        self.assertIn('adminPass. Value: None. None is not of type', str(ex))

    def test_change_password_adminpass_none(self):
        body = {'changePassword': None}
        self.assertRaises(self.validation_error,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    def test_change_password_bad_request(self):
        body = {'changePassword': {'pass': '12345'}}
        self.assertRaises(self.validation_error,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)

    def test_server_change_password_pass_disabled(self):
        # run with enable_instance_password disabled to verify adminPass
        # is missing from response. See lp bug 921814
        self.flags(enable_instance_password=False, group='api')
        body = {'changePassword': {'adminPass': '1234pass'}}
        res = self._get_action()(self.fake_req, fakes.FAKE_UUID, body=body)
        self._check_status(202, res, self._get_action())

    @mock.patch('nova.compute.api.API.set_admin_password',
                side_effect=exception.InstanceInvalidState(
                    instance_uuid='fake', attr='vm_state', state='stopped',
                    method='set_admin_password'))
    def test_change_password_invalid_state(self, mock_set_admin_password):
        body = {'changePassword': {'adminPass': 'test'}}
        self.assertRaises(webob.exc.HTTPConflict,
                          self._get_action(),
                          self.fake_req, fakes.FAKE_UUID, body=body)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Validators for ``quota`` namespaced extra specs."""

from nova.api.validation.extra_specs import base


EXTRA_SPEC_VALIDATORS = []


# CPU, memory, disk IO and VIF quotas (VMWare)
for key, name, unit in (
    ('cpu', 'CPU', 'MHz'),
    ('memory', 'memory', 'MB'),
    ('disk_io', 'disk IO', 'I/O per second'),
    ('vif', 'virtual interface', 'Mbps'),
):
    EXTRA_SPEC_VALIDATORS.extend(
        [
            base.ExtraSpecValidator(
                name=f'quota:{key}_limit',
                description=(
                    f'The upper limit for {name} allocation in {unit}. '
                    f'The utilization of an instance will not exceed this '
                    f'limit, even if there are available resources. '
                    f'This is typically used to ensure a consistent '
                    f'performance of instances independent of available '
                    f'resources.'
                    f'The value ``0`` indicates that {name} usage is not '
                    f'limited.'
                    f'Only supported by the VMWare virt driver.'
                ),
                value={
                    'type': int,
                    'min': 0,
                },
            ),
            base.ExtraSpecValidator(
                name=f'quota:{key}_reservation',
                description=(
                    f'The guaranteed minimum {name} reservation in {unit}. '
                    f'This means the specified amount of {name} that will '
                    f'be guaranteed for the instance. '
                    f'Only supported by the VMWare virt driver.'
                ),
                value={
                    'type': int,
                },
            ),
            base.ExtraSpecValidator(
                name=f'quota:{key}_shares_level',
                description=(
                    f"The allocation level for {name}. If you choose "
                    f"'custom', set the number of {name} shares using "
                    f"'quota:{key}_shares_share'. "
                    f"Only supported by the VMWare virt driver."
                ),
                value={
                    'type': str,
                    'enum': ['custom', 'high', 'normal', 'low'],
                },
            ),
            base.ExtraSpecValidator(
                name=f'quota:{key}_shares_share',
                description=(
                    f"The number of shares of {name} allocated in the "
                    f"event that 'quota:{key}_shares_level=custom' is "
                    f"used. "
                    f"Ignored otherwise. "
                    f"There is no unit for this value: it is a relative "
                    f"measure based on the settings for other instances. "
                    f"Only supported by the VMWare virt driver."
                ),
                value={
                    'type': int,
                    'min': 0,
                },
            ),
        ]
    )


# CPU quotas (libvirt)
EXTRA_SPEC_VALIDATORS.extend(
    [
        base.ExtraSpecValidator(
            name='quota:cpu_shares',
            description=(
                'The proportional weighted share for the domain. '
                'If this element is omitted, the service defaults to the OS '
                'provided defaults. '
                'There is no unit for the value; it is a relative measure '
                'based on the setting of other VMs. '
                'For example, a VM configured with a value of 2048 gets '
                'twice as much CPU time as a VM configured with value 1024. '
                'Only supported by the libvirt virt driver.'
            ),
            value={
                'type': int,
                'min': 0,
            },
        ),
        base.ExtraSpecValidator(
            name='quota:cpu_period',
            description=(
                'Specifies the enforcement interval in microseconds. '
                'Within a period, each VCPU of the instance is not allowed '
                'to consume more than the quota worth of runtime. '
                'The value should be in range 1,000 - 1,000,000. '
                'A period with a value of 0 means no value. '
                'Only supported by the libvirt virt driver.'
            ),
            value={
                'type': int,
                'min': 0,
            },
        ),
        base.ExtraSpecValidator(
            name='quota:cpu_quota',
            description=(
                "The maximum allowed bandwidth in microseconds. "
                "Can be combined with 'quota:cpu_period' to limit an instance "
                "to a percentage of capacity of a physical CPU. "
                "The value should be in range 1,000 - 2^64 or negative. "
                "A negative value indicates that the instance has infinite "
                "bandwidth. "
                "Only supported by the libvirt virt driver."
            ),
            value={
                'type': int,
            },
        ),
    ]
)


# Disk quotas (libvirt, HyperV)
for stat in ('read', 'write', 'total'):
    for metric in ('bytes', 'iops'):
        EXTRA_SPEC_VALIDATORS.append(
            base.ExtraSpecValidator(
                name=f'quota:disk_{stat}_{metric}_sec',
                # NOTE(stephenfin): HyperV supports disk_total_{metric}_sec
                # too; update
                description=(
                    f'The quota {stat} {metric} for disk. '
                    f'Only supported by the libvirt virt driver.'
                ),
                value={
                    'type': int,
                    'min': 0,
                },
            )
        )


# VIF quotas (libvirt)
# TODO(stephenfin): Determine whether this should be deprecated now that
# nova-network is dead
for stat in ('inbound', 'outbound'):
    for metric in ('average', 'peak', 'burst'):
        EXTRA_SPEC_VALIDATORS.append(
            base.ExtraSpecValidator(
                name=f'quota:vif_{stat}_{metric}',
                description=(
                    f'The quota {stat} {metric} for VIF. Only supported '
                    f'by the libvirt virt driver.'
                ),
                value={
                    'type': int,
                    'min': 0,
                },
            )
        )


def register():
    return EXTRA_SPEC_VALIDATORS



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova.objects import tag
from nova.tests.unit.objects import test_objects

RESOURCE_ID = '123'
TAG_NAME1 = 'fake-tag1'
TAG_NAME2 = 'fake-tag2'

fake_tag1 = {
    'resource_id': RESOURCE_ID,
    'tag': TAG_NAME1,
    }

fake_tag2 = {
    'resource_id': RESOURCE_ID,
    'tag': TAG_NAME1,
    }

fake_tag_list = [fake_tag1, fake_tag2]


def _get_tag(resource_id, tag_name, context=None):
    t = tag.Tag(context=context)
    t.resource_id = resource_id
    t.tag = tag_name
    return t


class _TestTagObject(object):
    @mock.patch('nova.db.main.api.instance_tag_add')
    def test_create(self, tag_add):
        tag_add.return_value = fake_tag1
        tag_obj = _get_tag(RESOURCE_ID, TAG_NAME1, context=self.context)
        tag_obj.create()

        tag_add.assert_called_once_with(self.context, RESOURCE_ID, TAG_NAME1)
        self.compare_obj(tag_obj, fake_tag1)

    @mock.patch('nova.db.main.api.instance_tag_delete')
    def test_destroy(self, tag_delete):
        tag.Tag.destroy(self.context, RESOURCE_ID, TAG_NAME1)
        tag_delete.assert_called_once_with(self.context,
                                           RESOURCE_ID, TAG_NAME1)

    @mock.patch('nova.db.main.api.instance_tag_exists')
    def test_exists(self, instance_tag_exists):
        tag.Tag.exists(self.context, RESOURCE_ID, TAG_NAME1)
        instance_tag_exists.assert_called_once_with(
            self.context, RESOURCE_ID, TAG_NAME1)


class TestMigrationObject(test_objects._LocalTest,
                          _TestTagObject):
    pass


class TestRemoteMigrationObject(test_objects._RemoteTest,
                                _TestTagObject):
    pass


class _TestTagList(object):
    def _compare_tag_list(self, tag_list, tag_list_obj):
        self.assertEqual(len(tag_list), len(tag_list_obj))

        for obj, fake in zip(tag_list_obj, tag_list):
            self.assertIsInstance(obj, tag.Tag)
            self.assertEqual(obj.tag, fake['tag'])
            self.assertEqual(obj.resource_id, fake['resource_id'])

    @mock.patch('nova.db.main.api.instance_tag_get_by_instance_uuid')
    def test_get_by_resource_id(self, get_by_inst):
        get_by_inst.return_value = fake_tag_list

        tag_list_obj = tag.TagList.get_by_resource_id(
            self.context, RESOURCE_ID)

        get_by_inst.assert_called_once_with(self.context, RESOURCE_ID)
        self._compare_tag_list(fake_tag_list, tag_list_obj)

    @mock.patch('nova.db.main.api.instance_tag_set')
    def test_create(self, tag_set):
        tag_set.return_value = fake_tag_list
        tag_list_obj = tag.TagList.create(
            self.context, RESOURCE_ID, [TAG_NAME1, TAG_NAME2])

        tag_set.assert_called_once_with(self.context,
                                        RESOURCE_ID, [TAG_NAME1, TAG_NAME2])
        self._compare_tag_list(fake_tag_list, tag_list_obj)

    @mock.patch('nova.db.main.api.instance_tag_delete_all')
    def test_destroy(self, tag_delete_all):
        tag.TagList.destroy(self.context, RESOURCE_ID)
        tag_delete_all.assert_called_once_with(self.context, RESOURCE_ID)


class TestTagList(test_objects._LocalTest, _TestTagList):
    pass


class TestTagListRemote(test_objects._RemoteTest, _TestTagList):
    pass



# Copyright 2015, 2018 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova import exception
from nova.i18n import _
from nova.virt.powervm.volume import fcvscsi


def build_volume_driver(adapter, instance, conn_info, stg_ftsk=None):
    drv_type = conn_info.get('driver_volume_type')
    if drv_type != 'fibre_channel':
        reason = _("Invalid connection type of %s") % drv_type
        raise exception.InvalidVolume(reason=reason)
    return fcvscsi.FCVscsiVolumeAdapter(adapter, instance, conn_info,
                                        stg_ftsk=stg_ftsk)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.tests import fixtures
from nova.tests.functional.notification_sample_tests \
    import notification_sample_base


class TestComputeTaskNotificationSample(
        notification_sample_base.NotificationSampleTestBase):

    def setUp(self):
        super(TestComputeTaskNotificationSample, self).setUp()
        self.neutron = fixtures.NeutronFixture(self)
        self.useFixture(self.neutron)

    def test_build_instances_fault(self):
        # Force down the compute node
        service_id = self.api.get_service_id('nova-compute')
        self.admin_api.put_service_force_down(service_id, True)

        server = self._boot_a_server(
            expected_status='ERROR',
            extra_params={'networks': [{'port': self.neutron.port_1['id']}]},
            additional_extra_specs={'hw:numa_nodes': 1,
                                    'hw:numa_cpus.0': '0',
                                    'hw:numa_mem.0': 512})
        self._wait_for_notification('compute_task.build_instances.error')
        # 0. scheduler.select_destinations.start
        # 1. compute_task.rebuild_server.error
        self.assertEqual(2, len(self.notifier.versioned_notifications),
                         self.notifier.versioned_notifications)
        self._verify_notification(
            'compute_task-build_instances-error',
            replacements={
                'instance_uuid': server['id'],
                'request_spec.instance_uuid': server['id'],
                'request_spec.security_groups': [],
                'request_spec.numa_topology.instance_uuid': server['id'],
                'request_spec.pci_requests.instance_uuid': server['id'],
                'reason.function_name': self.ANY,
                'reason.module_name': self.ANY,
                'reason.traceback': self.ANY
            },
            actual=self.notifier.versioned_notifications[1])

    def test_rebuild_fault(self):
        server = self._boot_a_server(
            extra_params={'networks': [{'port': self.neutron.port_1['id']}]},
            additional_extra_specs={'hw:numa_nodes': 1,
                                    'hw:numa_cpus.0': '0',
                                    'hw:numa_mem.0': 512})
        self._wait_for_notification('instance.create.end')
        # Force down the compute node
        service_id = self.api.get_service_id('nova-compute')
        self.admin_api.put_service_force_down(service_id, True)

        self.notifier.reset()

        # NOTE(takashin): The rebuild action and the evacuate action shares
        # same code path. So the 'evacuate' action is used for this test.
        self._evacuate_server(
            server, expected_state='ERROR', expected_migration_status='error')

        self._wait_for_notification('compute_task.rebuild_server.error')
        # 0. instance.evacuate
        # 1. scheduler.select_destinations.start
        # 2. compute_task.rebuild_server.error
        self.assertEqual(3, len(self.notifier.versioned_notifications),
                         self.notifier.versioned_notifications)
        self._verify_notification(
            'compute_task-rebuild_server-error',
            replacements={
                'instance_uuid': server['id'],
                'request_spec.instance_uuid': server['id'],
                'request_spec.security_groups': [],
                'request_spec.numa_topology.instance_uuid': server['id'],
                'request_spec.pci_requests.instance_uuid': server['id'],
                'reason.function_name': self.ANY,
                'reason.module_name': self.ANY,
                'reason.traceback': self.ANY
            },
            actual=self.notifier.versioned_notifications[2])

    def test_migrate_fault(self):
        server = self._boot_a_server(
            extra_params={'networks': [{'port': self.neutron.port_1['id']}]},
            additional_extra_specs={'hw:numa_nodes': 1,
                                    'hw:numa_cpus.0': '0',
                                    'hw:numa_mem.0': 512})
        self._wait_for_notification('instance.create.end')
        # Disable the compute node
        service_id = self.api.get_service_id('nova-compute')
        self.admin_api.put_service(service_id, {'status': 'disabled'})

        self.notifier.reset()

        # Note that the operation will return a 202 response but fail with
        # NoValidHost asynchronously.
        self.admin_api.post_server_action(server['id'], {'migrate': None})
        self._wait_for_notification('compute_task.migrate_server.error')
        self.assertEqual(1, len(self.notifier.versioned_notifications),
                         self.notifier.versioned_notifications)
        self._verify_notification(
            'compute_task-migrate_server-error',
            replacements={
                'instance_uuid': server['id'],
                'request_spec.instance_uuid': server['id'],
                'request_spec.security_groups': [],
                'request_spec.numa_topology.instance_uuid': server['id'],
                'request_spec.pci_requests.instance_uuid': server['id'],
                'reason.exception_message': 'No valid host was found. ',
                'reason.function_name': self.ANY,
                'reason.module_name': self.ANY,
                'reason.traceback': self.ANY
            },
            actual=self.notifier.versioned_notifications[0])



# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg

from nova import exception
from nova import objects
from nova.objects import fields
from nova.pci.request import PCI_REMOTE_MANAGED_TAG
from nova.pci import stats
from nova.pci import whitelist
from nova import test
from nova.tests.unit.pci import fakes

CONF = cfg.CONF
fake_pci_1 = {
    'compute_node_id': 1,
    'address': '0000:00:00.1',
    'product_id': 'p1',
    'vendor_id': 'v1',
    'status': 'available',
    'extra_k1': 'v1',
    'request_id': None,
    'numa_node': 0,
    'dev_type': fields.PciDeviceType.STANDARD,
    'parent_addr': None,
    }


fake_pci_2 = dict(fake_pci_1, vendor_id='v2',
                  product_id='p2',
                  address='0000:00:00.2',
                  numa_node=1)


fake_pci_3 = dict(fake_pci_1, address='0000:00:00.3')

fake_pci_4 = dict(fake_pci_1, vendor_id='v3',
                  product_id='p3',
                  address='0000:00:00.3',
                  numa_node= None)

pci_requests = [objects.InstancePCIRequest(count=1,
                    spec=[{'vendor_id': 'v1'}]),
                objects.InstancePCIRequest(count=1,
                    spec=[{'vendor_id': 'v2'}])]


pci_requests_multiple = [objects.InstancePCIRequest(count=1,
                             spec=[{'vendor_id': 'v1'}]),
                         objects.InstancePCIRequest(count=3,
                          spec=[{'vendor_id': 'v2'}])]


class PciDeviceStatsTestCase(test.NoDBTestCase):

    @staticmethod
    def _get_fake_requests(vendor_ids=None, numa_policy=None, count=1):
        if not vendor_ids:
            vendor_ids = ['v1', 'v2']

        specs = [{'vendor_id': vendor_id} for vendor_id in vendor_ids]

        return [objects.InstancePCIRequest(count=count, spec=[spec],
                numa_policy=numa_policy) for spec in specs]

    def _create_fake_devs(self):
        self.fake_dev_1 = objects.PciDevice.create(None, fake_pci_1)
        self.fake_dev_2 = objects.PciDevice.create(None, fake_pci_2)
        self.fake_dev_3 = objects.PciDevice.create(None, fake_pci_3)
        self.fake_dev_4 = objects.PciDevice.create(None, fake_pci_4)

        for dev in [self.fake_dev_1, self.fake_dev_2,
                    self.fake_dev_3, self.fake_dev_4]:
            self.pci_stats.add_device(dev)

    def _add_fake_devs_with_numa(self):
        fake_pci = dict(fake_pci_1, vendor_id='v4', product_id='pr4')
        devs = [dict(fake_pci, product_id='pr0', numa_node=0, ),
                dict(fake_pci, product_id='pr1', numa_node=1),
                dict(fake_pci, product_id='pr_none', numa_node=None)]

        for dev in devs:
            self.pci_stats.add_device(objects.PciDevice.create(None, dev))

    def setUp(self):
        super(PciDeviceStatsTestCase, self).setUp()
        self.pci_stats = stats.PciDeviceStats(objects.NUMATopology())
        # The following two calls need to be made before adding the devices.
        patcher = fakes.fake_pci_whitelist()
        self.addCleanup(patcher.stop)
        self._create_fake_devs()

    def test_add_device(self):
        self.assertEqual(len(self.pci_stats.pools), 3)
        self.assertEqual(set([d['vendor_id'] for d in self.pci_stats]),
                         set(['v1', 'v2', 'v3']))
        self.assertEqual(set([d['count'] for d in self.pci_stats]),
                         set([1, 2]))

    def test_remove_device(self):
        self.pci_stats.remove_device(self.fake_dev_2)
        self.assertEqual(len(self.pci_stats.pools), 2)
        self.assertEqual(self.pci_stats.pools[0]['count'], 2)
        self.assertEqual(self.pci_stats.pools[0]['vendor_id'], 'v1')

    def test_remove_device_exception(self):
        self.pci_stats.remove_device(self.fake_dev_2)
        self.assertRaises(exception.PciDevicePoolEmpty,
                          self.pci_stats.remove_device,
                          self.fake_dev_2)

    def test_pci_stats_equivalent(self):
        pci_stats2 = stats.PciDeviceStats(objects.NUMATopology())
        for dev in [self.fake_dev_1,
                    self.fake_dev_2,
                    self.fake_dev_3,
                    self.fake_dev_4]:
            pci_stats2.add_device(dev)
        self.assertEqual(self.pci_stats, pci_stats2)

    def test_pci_stats_not_equivalent(self):
        pci_stats2 = stats.PciDeviceStats(objects.NUMATopology())
        for dev in [self.fake_dev_1,
                    self.fake_dev_2,
                    self.fake_dev_3]:
            pci_stats2.add_device(dev)
        self.assertNotEqual(self.pci_stats, pci_stats2)

    def test_object_create(self):
        m = self.pci_stats.to_device_pools_obj()
        new_stats = stats.PciDeviceStats(objects.NUMATopology(), m)

        self.assertEqual(len(new_stats.pools), 3)
        self.assertEqual(set([d['count'] for d in new_stats]),
                         set([1, 2]))
        self.assertEqual(set([d['vendor_id'] for d in new_stats]),
                         set(['v1', 'v2', 'v3']))

    def test_apply_requests(self):
        self.assertEqual(len(self.pci_stats.pools), 3)
        self.pci_stats.apply_requests(pci_requests)
        self.assertEqual(len(self.pci_stats.pools), 2)
        self.assertEqual(self.pci_stats.pools[0]['vendor_id'], 'v1')
        self.assertEqual(self.pci_stats.pools[0]['count'], 1)

    def test_apply_requests_failed(self):
        self.assertRaises(exception.PciDeviceRequestFailed,
            self.pci_stats.apply_requests,
            pci_requests_multiple)

    def test_support_requests(self):
        self.assertTrue(self.pci_stats.support_requests(pci_requests))
        self.assertEqual(len(self.pci_stats.pools), 3)
        self.assertEqual(set([d['count'] for d in self.pci_stats]),
                         set((1, 2)))

    def test_support_requests_failed(self):
        self.assertFalse(
            self.pci_stats.support_requests(pci_requests_multiple))
        self.assertEqual(len(self.pci_stats.pools), 3)
        self.assertEqual(set([d['count'] for d in self.pci_stats]),
                         set([1, 2]))

    def test_support_requests_numa(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
            objects.InstanceNUMACell(
                id=1, cpuset=set(), pcpuset=set(), memory=0),
        ]
        self.assertTrue(self.pci_stats.support_requests(pci_requests, cells))

    def test_support_requests_numa_failed(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        self.assertFalse(self.pci_stats.support_requests(pci_requests, cells))

    def test_support_requests_no_numa_info(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_requests = self._get_fake_requests(vendor_ids=['v3'])
        self.assertTrue(self.pci_stats.support_requests(pci_requests, cells))

        # 'legacy' is the default numa_policy so the result must be same
        pci_requests = self._get_fake_requests(vendor_ids=['v3'],
            numa_policy = fields.PCINUMAAffinityPolicy.LEGACY)
        self.assertTrue(self.pci_stats.support_requests(pci_requests, cells))

    def test_support_requests_numa_pci_numa_policy_preferred(self):
        # numa node 0 has 2 devices with vendor_id 'v1'
        # numa node 1 has 1 device with vendor_id 'v2'
        # we request two devices with vendor_id 'v1' and 'v2'.
        # pci_numa_policy is 'preferred' so we can ignore numa affinity
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_requests = self._get_fake_requests(
            numa_policy=fields.PCINUMAAffinityPolicy.PREFERRED)

        self.assertTrue(self.pci_stats.support_requests(pci_requests, cells))

    def test_support_requests_no_numa_info_pci_numa_policy_required(self):
        # pci device with vendor_id 'v3' has numa_node=None.
        # pci_numa_policy is 'required' so we can't use this device
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_requests = self._get_fake_requests(vendor_ids=['v3'],
            numa_policy=fields.PCINUMAAffinityPolicy.REQUIRED)

        self.assertFalse(self.pci_stats.support_requests(pci_requests, cells))

    def test_filter_pools_for_socket_affinity_no_socket(self):
        self.pci_stats.numa_topology = objects.NUMATopology(
                cells=[objects.NUMACell(socket=None)])

        self.assertEqual(
            [],
            self.pci_stats._filter_pools_for_socket_affinity(
                self.pci_stats.pools, [objects.InstanceNUMACell()]))

    def test_filter_pools_for_socket_affinity(self):
        self.pci_stats.numa_topology = objects.NUMATopology(
                cells=[objects.NUMACell(id=1, socket=1)])

        pools = self.pci_stats._filter_pools_for_socket_affinity(
            self.pci_stats.pools, [objects.InstanceNUMACell(id=1)])
        self.assertEqual(1, len(pools))
        self.assertEqual('p2', pools[0]['product_id'])
        self.assertEqual('v2', pools[0]['vendor_id'])

    def test_consume_requests(self):
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['v1', 'v2']),
                         set([dev.vendor_id for dev in devs]))

    def test_consume_requests_empty(self):
        devs = self.pci_stats.consume_requests([])
        self.assertEqual(0, len(devs))

    def test_consume_requests_failed(self):
        self.assertIsNone(self.pci_stats.consume_requests(
                          pci_requests_multiple))

    def test_consume_requests_numa(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
            objects.InstanceNUMACell(
                id=1, cpuset=set(), pcpuset=set(), memory=0),
        ]
        devs = self.pci_stats.consume_requests(pci_requests, cells)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['v1', 'v2']),
                         set([dev.vendor_id for dev in devs]))

    def test_consume_requests_numa_failed(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        self.assertIsNone(self.pci_stats.consume_requests(pci_requests, cells))

    def test_consume_requests_no_numa_info(self):
        cells = [
            objects.InstanceNUMACell(
                id=0, cpuset=set(), pcpuset=set(), memory=0),
        ]
        pci_request = [objects.InstancePCIRequest(count=1,
                                                  spec=[{'vendor_id': 'v3'}])]
        devs = self.pci_stats.consume_requests(pci_request, cells)
        self.assertEqual(1, len(devs))
        self.assertEqual(set(['v3']),
                         set([dev.vendor_id for dev in devs]))

    def _test_consume_requests_numa_policy(self, cell_ids, policy,
            expected, vendor_id='v4', count=1):
        """Base test for 'consume_requests' function.

        Create three devices with vendor_id of 'v4': 'pr0' in NUMA node 0,
        'pr1' in NUMA node 1, 'pr_none' without NUMA affinity info. Attempt to
        consume a PCI request with a single device with ``vendor_id`` using the
        provided ``cell_ids`` and ``policy``. Compare result against
        ``expected``.
        """
        self._add_fake_devs_with_numa()
        cells = [
            objects.InstanceNUMACell(
                id=id, cpuset=set(), pcpuset=set(), memory=0)
            for id in cell_ids]

        pci_requests = self._get_fake_requests(vendor_ids=[vendor_id],
            numa_policy=policy, count=count)
        devs = self.pci_stats.consume_requests(pci_requests, cells)

        if expected is None:
            self.assertIsNone(devs)
        else:
            self.assertEqual(set(expected),
                             set([dev.product_id for dev in devs]))

    def test_consume_requests_numa_policy_required(self):
        """Ensure REQUIRED policy will ensure NUMA affinity.

        Policy is 'required' which means we must use a device with strict NUMA
        affinity. Request a device from NUMA node 0, which contains such a
        device, and ensure it's used.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.REQUIRED, ['pr0'])

    def test_consume_requests_numa_policy_required_fail(self):
        """Ensure REQUIRED policy will *only* provide NUMA affinity.

        Policy is 'required' which means we must use a device with strict NUMA
        affinity. Request a device from NUMA node 999, which does not contain
        any suitable devices, and ensure nothing is returned.
        """
        self._test_consume_requests_numa_policy(
            [999], fields.PCINUMAAffinityPolicy.REQUIRED, None)

    def test_consume_requests_numa_policy_legacy(self):
        """Ensure LEGACY policy will ensure NUMA affinity if possible.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request a device from NUMA node
        0, which contains such a device, and ensure it's used.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.LEGACY, ['pr0'])

    def test_consume_requests_numa_policy_legacy_fallback(self):
        """Ensure LEGACY policy will fallback to no NUMA affinity.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request a device from NUMA node
        999, which contains no such device, and ensure we fallback to the
        device without any NUMA affinity.
        """
        self._test_consume_requests_numa_policy(
            [999], fields.PCINUMAAffinityPolicy.LEGACY, ['pr_none'])

    def test_consume_requests_numa_policy_legacy_multiple(self):
        """Ensure LEGACY policy will use best policy for multiple devices.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request two devices from NUMA
        node 0, which contains only one such device, and ensure we use that
        device and the next best thing for the second device.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr0', 'pr_none'],
            count=2)

    def test_consume_requests_numa_policy_legacy_fail(self):
        """Ensure REQUIRED policy will *not* provide NUMA non-affinity.

        Policy is 'legacy' which means we must use a device with strict NUMA
        affinity or no provided NUMA affinity. Request a device with
        ``vendor_id`` of ``v2``, which can only be found in NUMA node 1, from
        NUMA node 0, and ensure nothing is returned.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.LEGACY, None, vendor_id='v2')

    def test_consume_requests_numa_policy_preferred(self):
        """Ensure PREFERRED policy will ensure NUMA affinity if possible.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request a device from NUMA node 0, which contains
        an affined device, and ensure it's used.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr0'])

    def test_consume_requests_numa_policy_preferred_fallback_a(self):
        """Ensure PREFERRED policy will fallback to no NUMA affinity.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request a device from NUMA node 999, which contains
        no such device, and ensure we fallback to the device without any NUMA
        affinity.
        """
        self._test_consume_requests_numa_policy(
            [999], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr_none'])

    def test_consume_requests_numa_policy_preferred_fallback_b(self):
        """Ensure PREFERRED policy will fallback to different NUMA affinity.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request a device with ``vendor_id`` of ``v2``, which
        can only be found in NUMA node 1, from NUMA node 0, and ensure we
        fallback to this device.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['p2'],
            vendor_id='v2')

    def test_consume_requests_numa_policy_preferred_multiple_a(self):
        """Ensure PREFERRED policy will use best policy for multiple devices.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request two devices from NUMA node 0, which contains
        only one such device, and ensure we use that device and gracefully
        degrade for the other device.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED, ['pr0', 'pr_none'],
            count=2)

    def test_consume_requests_numa_policy_preferred_multiple_b(self):
        """Ensure PREFERRED policy will use best policy for multiple devices.

        Policy is 'preferred' which means we must use a device with any level
        of NUMA affinity. Request three devices from NUMA node 0, which
        contains only one such device, and ensure we use that device and
        gracefully degrade for the other devices.
        """
        self._test_consume_requests_numa_policy(
            [0], fields.PCINUMAAffinityPolicy.PREFERRED,
            ['pr0', 'pr_none', 'pr1'], count=3)

    @mock.patch(
        'nova.pci.whitelist.Whitelist._parse_white_list_from_config')
    def test_white_list_parsing(self, mock_whitelist_parse):
        white_list = '{"product_id":"0001", "vendor_id":"8086"}'
        CONF.set_override('passthrough_whitelist', white_list, 'pci')
        pci_stats = stats.PciDeviceStats(objects.NUMATopology())
        pci_stats.add_device(self.fake_dev_2)
        pci_stats.remove_device(self.fake_dev_2)
        self.assertEqual(1, mock_whitelist_parse.call_count)


class PciDeviceStatsWithTagsTestCase(test.NoDBTestCase):

    def setUp(self):
        super(PciDeviceStatsWithTagsTestCase, self).setUp()
        white_list = ['{"vendor_id":"1137","product_id":"0071",'
                        '"address":"*:0a:00.*","physical_network":"physnet1"}',
                      '{"vendor_id":"1137","product_id":"0072"}',
                      '{"vendor_id":"15b3","product_id":"101e", '
                      '"remote_managed": "true"}',
                      '{"vendor_id":"15b3","product_id":"101c"}',
                      '{"vendor_id":"15b3","product_id":"1018", '
                      '"remote_managed": "false"}']
        self.flags(passthrough_whitelist=white_list, group='pci')
        dev_filter = whitelist.Whitelist(white_list)
        self.pci_stats = stats.PciDeviceStats(
            objects.NUMATopology(),
            dev_filter=dev_filter)

    def _create_pci_devices(self):
        self.pci_tagged_devices = []
        for dev in range(4):
            pci_dev = {'compute_node_id': 1,
                       'address': '0000:0a:00.%d' % dev,
                       'vendor_id': '1137',
                       'product_id': '0071',
                       'status': 'available',
                       'request_id': None,
                       'dev_type': 'type-PCI',
                       'parent_addr': None,
                       'numa_node': 0}
            self.pci_tagged_devices.append(objects.PciDevice.create(None,
                                                                    pci_dev))

        self.pci_untagged_devices = []
        for dev in range(3):
            pci_dev = {'compute_node_id': 1,
                       'address': '0000:0b:00.%d' % dev,
                       'vendor_id': '1137',
                       'product_id': '0072',
                       'status': 'available',
                       'request_id': None,
                       'dev_type': 'type-PCI',
                       'parent_addr': None,
                       'numa_node': 0}
            self.pci_untagged_devices.append(objects.PciDevice.create(None,
                                                                      pci_dev))

        self.locally_managed_netdevs = []
        self.remote_managed_netdevs = []
        self.remote_managed_netdevs.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:0c:00.1',
                    'vendor_id': '15b3',
                    'product_id': '101e',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:0c:00.0',
                    'numa_node': 0,
                    "capabilities": {"vpd": {
                        "card_serial_number": "MT2113X00000"}}
                }))

        # For testing implicit remote_managed == False tagging.
        self.locally_managed_netdevs.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:0d:00.1',
                    'vendor_id': '15b3',
                    'product_id': '101c',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:0d:00.0',
                    'numa_node': 0}))

        # For testing explicit remote_managed == False tagging.
        self.locally_managed_netdevs.append(
            objects.PciDevice.create(
                None, {
                    'compute_node_id': 1,
                    'address': '0000:0e:00.1',
                    'vendor_id': '15b3',
                    'product_id': '1018',
                    'status': 'available',
                    'request_id': None,
                    'dev_type': fields.PciDeviceType.SRIOV_VF,
                    'parent_addr': '0000:0e:00.0',
                    'numa_node': 0}))

        for dev in self.pci_tagged_devices:
            self.pci_stats.add_device(dev)

        for dev in self.pci_untagged_devices:
            self.pci_stats.add_device(dev)

        for dev in self.remote_managed_netdevs:
            self.pci_stats.add_device(dev)

        for dev in self.locally_managed_netdevs:
            self.pci_stats.add_device(dev)

    def _assertPoolContent(self, pool, vendor_id, product_id, count, **tags):
        self.assertEqual(vendor_id, pool['vendor_id'])
        self.assertEqual(product_id, pool['product_id'])
        self.assertEqual(count, pool['count'])
        if tags:
            for k, v in tags.items():
                self.assertEqual(v, pool[k])

    def _assertPools(self):
        # Pools are ordered based on the number of keys. 'product_id',
        # 'vendor_id' are always part of the keys. When tags are present,
        # they are also part of the keys. In this test class, we have
        # 5 pools with the second one having the tag 'physical_network'
        # and the value 'physnet1' and multiple pools for testing
        # variations of explicit/implicit remote_managed tagging.
        self.assertEqual(5, len(self.pci_stats.pools))
        self._assertPoolContent(self.pci_stats.pools[0], '1137', '0072',
                                len(self.pci_untagged_devices))
        self.assertEqual(self.pci_untagged_devices,
                         self.pci_stats.pools[0]['devices'])
        self._assertPoolContent(self.pci_stats.pools[1], '1137', '0071',
                                len(self.pci_tagged_devices),
                                physical_network='physnet1')
        self.assertEqual(self.pci_tagged_devices,
                         self.pci_stats.pools[1]['devices'])
        self._assertPoolContent(self.pci_stats.pools[2], '15b3', '101e',
                                len(self.remote_managed_netdevs),
                                remote_managed='true')
        self.assertEqual(self.remote_managed_netdevs,
                         self.pci_stats.pools[2]['devices'])
        self._assertPoolContent(self.pci_stats.pools[3], '15b3', '101c', 1,
                                remote_managed='false')
        self.assertEqual([self.locally_managed_netdevs[0]],
                         self.pci_stats.pools[3]['devices'])
        self._assertPoolContent(self.pci_stats.pools[4], '15b3', '1018', 1,
                                remote_managed='false')
        self.assertEqual([self.locally_managed_netdevs[1]],
                         self.pci_stats.pools[4]['devices'])

    def test_add_devices(self):
        self._create_pci_devices()
        self._assertPools()

    def test_consume_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=1,
                            spec=[{'physical_network': 'physnet1'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '1137',
                                   'product_id': '0072'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b3',
                                   'product_id': '101e',
                                   PCI_REMOTE_MANAGED_TAG: 'True'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b3',
                                   'product_id': '101c',
                                   PCI_REMOTE_MANAGED_TAG: 'False'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'vendor_id': '15b3',
                                   'product_id': '1018',
                                   PCI_REMOTE_MANAGED_TAG: 'False'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(5, len(devs))
        self.assertEqual(set(['0071', '0072', '1018', '101e', '101c']),
                         set([dev.product_id for dev in devs]))
        self._assertPoolContent(self.pci_stats.pools[0], '1137', '0072', 2)
        self._assertPoolContent(self.pci_stats.pools[1], '1137', '0071', 3,
                                physical_network='physnet1')
        self._assertPoolContent(self.pci_stats.pools[2], '15b3', '101e', 0,
                                remote_managed='true')
        self._assertPoolContent(self.pci_stats.pools[3], '15b3', '101c', 0,
                                remote_managed='false')
        self._assertPoolContent(self.pci_stats.pools[4], '15b3', '1018', 0,
                                remote_managed='false')

    def test_add_device_no_devspec(self):
        self._create_pci_devices()
        pci_dev = {'compute_node_id': 1,
                   'address': '0000:0c:00.1',
                   'vendor_id': '2345',
                   'product_id': '0172',
                   'status': 'available',
                   'parent_addr': None,
                   'request_id': None}
        pci_dev_obj = objects.PciDevice.create(None, pci_dev)
        self.pci_stats.add_device(pci_dev_obj)
        # There should be no change
        self.assertIsNone(
            self.pci_stats._create_pool_keys_from_dev(pci_dev_obj))
        self._assertPools()

    def test_remove_device_no_devspec(self):
        self._create_pci_devices()
        pci_dev = {'compute_node_id': 1,
                   'address': '0000:0c:00.1',
                   'vendor_id': '2345',
                   'product_id': '0172',
                   'status': 'available',
                   'parent_addr': None,
                   'request_id': None}
        pci_dev_obj = objects.PciDevice.create(None, pci_dev)
        self.pci_stats.remove_device(pci_dev_obj)
        # There should be no change
        self.assertIsNone(
            self.pci_stats._create_pool_keys_from_dev(pci_dev_obj))
        self._assertPools()

    def test_remove_device(self):
        self._create_pci_devices()
        dev1 = self.pci_untagged_devices.pop()
        self.pci_stats.remove_device(dev1)
        dev2 = self.pci_tagged_devices.pop()
        self.pci_stats.remove_device(dev2)
        self._assertPools()

    def test_update_device(self):
        # Update device type of one of the device from type-PCI to
        # type-PF. Verify if the existing pool is updated and a new
        # pool is created with dev_type type-PF.
        self._create_pci_devices()
        dev1 = self.pci_tagged_devices.pop()
        dev1.dev_type = 'type-PF'
        self.pci_stats.update_device(dev1)
        self.assertEqual(6, len(self.pci_stats.pools))
        self._assertPoolContent(self.pci_stats.pools[0], '1137', '0072',
                                len(self.pci_untagged_devices))
        self.assertEqual(self.pci_untagged_devices,
                         self.pci_stats.pools[0]['devices'])
        self._assertPoolContent(self.pci_stats.pools[1], '1137', '0071',
                                len(self.pci_tagged_devices),
                                physical_network='physnet1')
        self.assertEqual(self.pci_tagged_devices,
                         self.pci_stats.pools[1]['devices'])
        self._assertPoolContent(self.pci_stats.pools[5], '1137', '0071',
                                1,
                                physical_network='physnet1',
                                remote_managed='false')
        self.assertEqual(dev1,
                         self.pci_stats.pools[5]['devices'][0])


class PciDeviceVFPFStatsTestCase(test.NoDBTestCase):

    def setUp(self):
        super(PciDeviceVFPFStatsTestCase, self).setUp()
        white_list = ['{"vendor_id":"8086","product_id":"1528"}',
                      '{"vendor_id":"8086","product_id":"1515"}',
                      '{"vendor_id":"15b3","product_id":"a2d6", '
                      '"remote_managed": "false"}',
                      '{"vendor_id":"15b3","product_id":"101e", '
                      '"remote_managed": "true"}']
        self.flags(passthrough_whitelist=white_list, group='pci')
        self.pci_stats = stats.PciDeviceStats(objects.NUMATopology())

    def _create_pci_devices(self, vf_product_id=1515, pf_product_id=1528):
        self.sriov_pf_devices = []
        for dev in range(2):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0000:81:00.%d' % dev,
                'vendor_id': '8086',
                'product_id': '%d' % pf_product_id,
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_PF,
                'parent_addr': None,
                'numa_node': 0
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.child_devices = []
            self.sriov_pf_devices.append(dev_obj)

        # PF devices for remote_managed VFs.
        self.sriov_pf_devices_remote = []
        for dev in range(2):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0001:81:00.%d' % dev,
                'vendor_id': '15b3',
                'product_id': 'a2d6',
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_PF,
                'parent_addr': None,
                'numa_node': 0,
                "capabilities": {"vpd": {
                    "card_serial_number": "MT2113X00000"}},
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.child_devices = []
            self.sriov_pf_devices_remote.append(dev_obj)

        self.sriov_vf_devices = []
        for dev in range(8):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0000:81:10.%d' % dev,
                'vendor_id': '8086',
                'product_id': '%d' % vf_product_id,
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_VF,
                'parent_addr': '0000:81:00.%d' % int(dev / 4),
                'numa_node': 0
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.parent_device = self.sriov_pf_devices[int(dev / 4)]
            dev_obj.parent_device.child_devices.append(dev_obj)
            self.sriov_vf_devices.append(dev_obj)

        self.sriov_vf_devices_remote = []
        for dev in range(8):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0001:81:10.%d' % dev,
                'vendor_id': '15b3',
                'product_id': '101e',
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.SRIOV_VF,
                'parent_addr': '0001:81:00.%d' % int(dev / 4),
                'numa_node': 0,
                "capabilities": {"vpd": {"card_serial_number": "MT2113X00000"}}
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.parent_device = self.sriov_pf_devices_remote[int(dev / 4)]
            dev_obj.parent_device.child_devices.append(dev_obj)
            self.sriov_vf_devices_remote.append(dev_obj)

        self.vdpa_devices = []
        for dev in range(8):
            pci_dev = {
                'compute_node_id': 1,
                'address': '0000:82:10.%d' % dev,
                'vendor_id': '8086',
                'product_id': '%d' % vf_product_id,
                'status': 'available',
                'request_id': None,
                'dev_type': fields.PciDeviceType.VDPA,
                'parent_addr': '0000:81:00.%d' % int(dev / 4),
                'numa_node': 0
            }
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.parent_device = self.sriov_pf_devices[int(dev / 4)]
            dev_obj.parent_device.child_devices.append(dev_obj)
            self.vdpa_devices.append(dev_obj)

        list(map(self.pci_stats.add_device, self.sriov_pf_devices))
        list(map(self.pci_stats.add_device, self.sriov_vf_devices))
        list(map(self.pci_stats.add_device, self.vdpa_devices))
        list(map(self.pci_stats.add_device, self.sriov_pf_devices_remote))
        list(map(self.pci_stats.add_device, self.sriov_vf_devices_remote))

    def test_consume_VDPA_requests(self):
        self._create_pci_devices()
        pci_requests = [
            objects.InstancePCIRequest(
                count=8, spec=[{'dev_type': 'vdpa'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(8, len(devs))
        self.assertEqual('vdpa', devs[0].dev_type)
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these devs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])

    def test_consume_VF_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1515'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['1515']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these VFs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])

    def test_consume_PF_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1528',
                                    'dev_type': 'type-PF'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['1528']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that there are no free devices left, as when allocating
        # both available PFs, its VFs should not be available.
        self.assertEqual(0, len([d for d in free_devs
                                 if d.product_id == '1515']))

    def test_consume_VF_and_PF_requests(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1515'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '1528',
                                    'dev_type': 'type-PF'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(3, len(devs))
        self.assertEqual(set(['1528', '1515']),
                            set([dev.product_id for dev in devs]))

    def test_consume_VF_and_PF_requests_failed(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=5,
                            spec=[{'product_id': '1515'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '1528',
                                    'dev_type': 'type-PF'}])]
        self.assertIsNone(self.pci_stats.consume_requests(pci_requests))

    def test_consume_VF_and_PF_same_prodict_id_failed(self):
        self._create_pci_devices(pf_product_id=1515)
        pci_requests = [objects.InstancePCIRequest(count=9,
                            spec=[{'product_id': '1515'}])]
        self.assertIsNone(self.pci_stats.consume_requests(pci_requests))

    def test_consume_PF_not_remote_managed(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{'product_id': '1528',
                                   'dev_type': 'type-PF',
                                   PCI_REMOTE_MANAGED_TAG: 'false'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['1528']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that there are no free devices left with the
        # product ID under test, as when allocating both available
        # PFs, its VFs should not be available.
        self.assertEqual(0, len([d for d in free_devs
                                 if d.product_id == '1528']))

    def test_consume_VF_requests_remote_managed(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=2,
                            spec=[{PCI_REMOTE_MANAGED_TAG: 'true'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['101e']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these VFs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])

    def test_consume_VF_requests_remote_managed_filtered(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '101e',
                                   PCI_REMOTE_MANAGED_TAG: 'false'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '101e'}])]
        free_devs_before = self.pci_stats.get_free_devs()
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertIsNone(devs)
        free_devs_after = self.pci_stats.get_free_devs()
        self.assertEqual(free_devs_before, free_devs_after)

    def test_consume_VF_requests_remote_managed_mix(self):
        self._create_pci_devices()
        pci_requests = [objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '101e',
                                   PCI_REMOTE_MANAGED_TAG: 'true'}]),
                        objects.InstancePCIRequest(count=1,
                            spec=[{'product_id': '1515',
                                   PCI_REMOTE_MANAGED_TAG: 'false'}])]
        devs = self.pci_stats.consume_requests(pci_requests)
        self.assertEqual(2, len(devs))
        self.assertEqual(set(['101e', '1515']),
                            set([dev.product_id for dev in devs]))
        free_devs = self.pci_stats.get_free_devs()
        # Validate that the parents of these VFs has been removed
        # from pools.
        for dev in devs:
            self.assertNotIn(dev.parent_addr,
                             [free_dev.address for free_dev in free_devs])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from .api_paste import ApiPasteNoProjectId  # noqa: F401
from .api_paste import ApiPasteV21Fixture  # noqa: F401
from .cast_as_call import CastAsCallFixture  # noqa: F401
from .cinder import CinderFixture  # noqa: F401
from .conf import ConfFixture  # noqa: F401, F403
from .cyborg import CyborgFixture  # noqa: F401
from .glance import GlanceFixture  # noqa: F401
from .libvirt import LibvirtFixture  # noqa: F401
from .libvirt_imagebackend import LibvirtImageBackendFixture  # noqa: F401
from .neutron import NeutronFixture  # noqa: F401
from .notifications import NotificationFixture  # noqa: F401
from .nova import *  # noqa: F401, F403
from .os_brick import OSBrickFixture  # noqa: F401
from .policy import OverridePolicyFixture  # noqa: F401
from .policy import PolicyFixture  # noqa: F401
from .policy import RealPolicyFixture  # noqa: F401
from .policy import RoleBasedPolicyFixture  # noqa: F401



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import functools
import threading

import eventlet
import fixtures
from oslo_log import log as logging
import oslo_messaging
from oslo_serialization import jsonutils
from oslo_utils import excutils
from oslo_utils import timeutils

from nova import rpc

LOG = logging.getLogger(__name__)


class _Sub(object):
    """Allow a subscriber to efficiently wait for an event to occur, and
    retrieve events which have occurred.
    """

    def __init__(self):
        self._cond = threading.Condition()
        self._notifications = []

    def received(self, notification):
        with self._cond:
            self._notifications.append(notification)
            self._cond.notifyAll()

    def wait_n(self, n, event, timeout):
        """Wait until at least n notifications have been received, and return
        them. May return less than n notifications if timeout is reached.
        """

        with timeutils.StopWatch(timeout) as timer:
            with self._cond:
                while len(self._notifications) < n:
                    if timer.expired():
                        # notifications = pprint.pformat(
                        #     {event: sub._notifications
                        #      for event, sub in VERSIONED_SUBS.items()})
                        # FIXME: tranform this to get access to all the
                        # versioned notifications
                        notifications = []
                        raise AssertionError(
                            "Notification %(event)s hasn't been "
                            "received. Received:\n%(notifications)s" % {
                                'event': event,
                                'notifications': notifications,
                            })
                    self._cond.wait(timer.leftover())

                # Return a copy of the notifications list
                return list(self._notifications)


FakeMessage = collections.namedtuple(
    'FakeMessage',
    ['publisher_id', 'priority', 'event_type', 'payload', 'context'])


class FakeNotifier(object):

    def __init__(
        self, transport, publisher_id, serializer=None, parent=None,
        test_case_id=None
    ):
        self.transport = transport
        self.publisher_id = publisher_id
        self._serializer = \
            serializer or oslo_messaging.serializer.NoOpSerializer()
        if parent:
            self.notifications = parent.notifications
        else:
            self.notifications = []

        for priority in ['debug', 'info', 'warn', 'error', 'critical']:
            setattr(
                self, priority,
                functools.partial(self._notify, priority.upper()),
            )

        self.test_case_id = test_case_id

    def prepare(self, publisher_id=None):
        if publisher_id is None:
            publisher_id = self.publisher_id

        return self.__class__(
            self.transport, publisher_id,
            serializer=self._serializer, parent=self,
            test_case_id=self.test_case_id
        )

    def _notify(self, priority, ctxt, event_type, payload):
        try:
            payload = self._serializer.serialize_entity(ctxt, payload)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error('Error serializing payload: %s', payload)

        # NOTE(sileht): simulate the kombu serializer
        # this permit to raise an exception if something have not
        # been serialized correctly
        jsonutils.to_primitive(payload)
        # NOTE(melwitt): Try to serialize the context, as the rpc would.
        #                An exception will be raised if something is wrong
        #                with the context.
        self._serializer.serialize_context(ctxt)
        msg = FakeMessage(
            self.publisher_id, priority, event_type, payload, ctxt)
        self.notifications.append(msg)

    def is_enabled(self):
        return True

    def reset(self):
        self.notifications.clear()


class FakeVersionedNotifier(FakeNotifier):
    def __init__(
        self, transport, publisher_id, serializer=None, parent=None,
        test_case_id=None
    ):
        super().__init__(
            transport, publisher_id, serializer, test_case_id=test_case_id)
        if parent:
            self.versioned_notifications = parent.versioned_notifications
        else:
            self.versioned_notifications = []

        if parent:
            self.subscriptions = parent.subscriptions
        else:
            self.subscriptions = collections.defaultdict(_Sub)

    @staticmethod
    def _get_sender_test_case_id():
        current = eventlet.getcurrent()
        # NOTE(gibi) not all eventlet spawn is under our control, so there can
        # be senders without test_case_id set, find the first ancestor that
        # was spawned from nova.utils.spawn[_n] and therefore has the id set.
        while not getattr(current, 'test_case_id', None):
            current = current.parent
        return current.test_case_id

    def _notify(self, priority, ctxt, event_type, payload):
        sender_test_case_id = self._get_sender_test_case_id()
        # NOTE(gibi): this is here to prevent late notifications from already
        # finished test cases to break the currently running test case. See
        # more in https://bugs.launchpad.net/nova/+bug/1946339
        if sender_test_case_id != self.test_case_id:
            raise RuntimeError(
                'FakeVersionedNotifier received %s notification emitted by %s '
                'test case which is different from the currently running test '
                'case %s. This notification is ignored. The sender test case '
                'probably leaked a running eventlet that emitted '
                'notifications after the test case finished. Now this '
                'eventlet is terminated by raising this exception.' %
                (event_type, sender_test_case_id, self.test_case_id))

        payload = self._serializer.serialize_entity(ctxt, payload)
        notification = {
            'publisher_id': self.publisher_id,
            'priority': priority,
            'event_type': event_type,
            'payload': payload,
        }
        self.versioned_notifications.append(notification)
        self.subscriptions[event_type].received(notification)

    def reset(self):
        self.versioned_notifications.clear()
        self.subscriptions.clear()

    def wait_for_versioned_notifications(
        self, event_type, n_events=1, timeout=10.0,
    ):
        return self.subscriptions[event_type].wait_n(
            n_events, event_type, timeout)


class NotificationFixture(fixtures.Fixture):
    def __init__(self, test):
        self.test = test

    def setUp(self):
        super().setUp()
        self.addCleanup(self.reset)

        self.fake_notifier = FakeNotifier(
            rpc.LEGACY_NOTIFIER.transport,
            rpc.LEGACY_NOTIFIER.publisher_id,
            serializer=getattr(
                rpc.LEGACY_NOTIFIER, '_serializer', None))
        self.fake_versioned_notifier = FakeVersionedNotifier(
            rpc.NOTIFIER.transport,
            rpc.NOTIFIER.publisher_id,
            serializer=getattr(rpc.NOTIFIER, '_serializer', None),
            test_case_id=self.test.id()
        )
        if rpc.LEGACY_NOTIFIER and rpc.NOTIFIER:
            self.test.stub_out('nova.rpc.LEGACY_NOTIFIER', self.fake_notifier)
            self.test.stub_out(
                'nova.rpc.NOTIFIER', self.fake_versioned_notifier)

    def reset(self):
        self.fake_notifier.reset()
        self.fake_versioned_notifier.reset()

    def wait_for_versioned_notifications(
        self, event_type, n_events=1, timeout=10.0,
    ):
        return self.fake_versioned_notifier.wait_for_versioned_notifications(
            event_type, n_events, timeout,
        )

    @property
    def versioned_notifications(self):
        return self.fake_versioned_notifier.versioned_notifications

    @property
    def notifications(self):
        return self.fake_notifier.notifications



# Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import mock

from nova import test
from nova import utils
from nova.virt.libvirt.volume import fs

FAKE_MOUNT_POINT = '/var/lib/nova/fake-mount'
FAKE_SHARE = 'fake-share'
NORMALIZED_SHARE = FAKE_SHARE + '-normalized'
HASHED_SHARE = utils.get_hash_str(NORMALIZED_SHARE)
FAKE_DEVICE_NAME = 'fake-device'


class FakeFileSystemVolumeDriver(fs.LibvirtBaseFileSystemVolumeDriver):

    def _get_mount_point_base(self):
        return FAKE_MOUNT_POINT

    def _normalize_export(self, export):
        return NORMALIZED_SHARE


class LibvirtBaseFileSystemVolumeDriverTestCase(test.NoDBTestCase):
    """Tests the basic behavior of the LibvirtBaseFileSystemVolumeDriver"""

    def setUp(self):
        super(LibvirtBaseFileSystemVolumeDriverTestCase, self).setUp()
        self.connection = mock.Mock()
        self.driver = FakeFileSystemVolumeDriver(self.connection)
        self.connection_info = {
            'data': {
                'export': FAKE_SHARE,
                'name': FAKE_DEVICE_NAME,
            }
        }

    def test_get_device_path(self):
        path = self.driver._get_device_path(self.connection_info)
        expected_path = os.path.join(FAKE_MOUNT_POINT,
                                     HASHED_SHARE,
                                     FAKE_DEVICE_NAME)
        self.assertEqual(expected_path, path)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova.api.openstack.compute import hypervisors
from nova.policies import base as base_policy
from nova.policies import hypervisors as hv_policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.policies import base


class HypervisorsPolicyTest(base.BasePolicyTest):
    """Test os-hypervisors APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(HypervisorsPolicyTest, self).setUp()
        self.controller = hypervisors.HypervisorsController()
        self.req = fakes.HTTPRequest.blank('')
        self.controller._get_compute_nodes_by_name_pattern = mock.MagicMock()
        self.controller.host_api.compute_node_get_all = mock.MagicMock()
        self.controller.host_api.service_get_by_compute_host = mock.MagicMock()
        self.controller.host_api.compute_node_get = mock.MagicMock()

        # With legacy rule and scope check disabled by default, system admin,
        # legacy admin, and project admin will be able to perform hypervisors
        # Operations.
        self.system_admin_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context]

    def test_list_hypervisors_policy(self):
        rule_name = hv_policies.BASE_POLICY_NAME % 'list'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.index,
                                self.req)

    def test_list_details_hypervisors_policy(self):
        rule_name = hv_policies.BASE_POLICY_NAME % 'list-detail'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.detail,
                                self.req)

    def test_show_hypervisors_policy(self):
        rule_name = hv_policies.BASE_POLICY_NAME % 'show'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.show,
                                self.req, 11111)

    @mock.patch('nova.compute.api.HostAPI.get_host_uptime')
    def test_uptime_hypervisors_policy(self, mock_uptime):
        rule_name = hv_policies.BASE_POLICY_NAME % 'uptime'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.uptime,
                                self.req, 11111)

    def test_search_hypervisors_policy(self):
        rule_name = hv_policies.BASE_POLICY_NAME % 'search'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.search,
                                self.req, 11111)

    def test_servers_hypervisors_policy(self):
        rule_name = hv_policies.BASE_POLICY_NAME % 'servers'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.servers,
                                self.req, 11111)

    @mock.patch('nova.compute.api.HostAPI.compute_node_statistics')
    def test_statistics_hypervisors_policy(self, mock_statistics):
        rule_name = hv_policies.BASE_POLICY_NAME % 'statistics'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.statistics,
                                self.req)


class HypervisorsNoLegacyNoScopePolicyTest(HypervisorsPolicyTest):
    """Test Hypervisors APIs policies with no legacy deprecated rules
    and no scope checks which means new defaults only. In this case
    system admin, legacy admin, and project admin will be able to perform
    Hypervisors Operations. Legacy admin will be allowed as policy is just
    admin if no scope checks.
    """

    without_deprecated_rules = True


class HypervisorsScopeTypePolicyTest(HypervisorsPolicyTest):
    """Test os-hypervisors APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(HypervisorsScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # With scope checks enable, only system admin is able to perform
        # hypervisors Operations.
        self.system_admin_authorized_contexts = [self.system_admin_context]


class HypervisorsScopeTypeNoLegacyPolicyTest(HypervisorsScopeTypePolicyTest):
    """Test Hypervisors APIs policies with no legacy deprecated rules
    and scope checks enabled which means scope + new defaults so
    only system admin is able to perform hypervisors Operations.
    """

    without_deprecated_rules = True

    rules_without_deprecation = {
        hv_policies.BASE_POLICY_NAME % 'list':
            base_policy.ADMIN,
        hv_policies.BASE_POLICY_NAME % 'list-detail':
            base_policy.ADMIN,
        hv_policies.BASE_POLICY_NAME % 'show':
            base_policy.ADMIN,
        hv_policies.BASE_POLICY_NAME % 'statistics':
            base_policy.ADMIN,
        hv_policies.BASE_POLICY_NAME % 'uptime':
            base_policy.ADMIN,
        hv_policies.BASE_POLICY_NAME % 'search':
            base_policy.ADMIN,
        hv_policies.BASE_POLICY_NAME % 'servers':
            base_policy.ADMIN,
    }



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

import mock
from webob import exc

from nova.api.openstack import api_version_request as api_version
from nova.api.openstack.compute import flavor_access \
        as flavor_access_v21
from nova.api.openstack.compute import flavors as flavors_api
from nova import context
from nova import exception
from nova import test
from nova.tests.unit.api.openstack import fakes


def generate_flavor(flavorid, ispublic):
    return {
        'id': flavorid,
        'flavorid': str(flavorid),
        'root_gb': 1,
        'ephemeral_gb': 1,
        'name': u'test',
        'created_at': datetime.datetime(2012, 1, 1, 1, 1, 1, 1),
        'updated_at': None,
        'memory_mb': 512,
        'vcpus': 1,
        'swap': 512,
        'rxtx_factor': 1.0,
        'disabled': False,
        'extra_specs': {},
        'vcpu_weight': None,
        'is_public': bool(ispublic),
        'description': None
    }


FLAVORS = {
    '0': generate_flavor(0, True),
    '1': generate_flavor(1, True),
    '2': generate_flavor(2, False),
    '3': generate_flavor(3, False)}


ACCESS_LIST = [
    {'flavor_id': '2', 'project_id': 'proj2'},
    {'flavor_id': '2', 'project_id': 'proj3'},
    {'flavor_id': '3', 'project_id': 'proj3'},
]


def fake_get_flavor_access_by_flavor_id(context, flavorid):
    res = []
    for access in ACCESS_LIST:
        if access['flavor_id'] == flavorid:
            res.append(access['project_id'])
    return res


def fake_get_flavor_by_flavor_id(context, flavorid):
    return FLAVORS[flavorid]


def _has_flavor_access(flavorid, projectid):
    for access in ACCESS_LIST:
        if (access['flavor_id'] == flavorid and
                access['project_id'] == projectid):
            return True
    return False


def fake_get_all_flavors_sorted_list(context, inactive=False,
                                     filters=None, sort_key='flavorid',
                                     sort_dir='asc', limit=None, marker=None):
    if filters is None or filters['is_public'] is None:
        return sorted(FLAVORS.values(), key=lambda item: item[sort_key])

    res = {}
    for k, v in FLAVORS.items():
        if filters['is_public'] and _has_flavor_access(k, context.project_id):
            res.update({k: v})
            continue
        if v['is_public'] == filters['is_public']:
            res.update({k: v})

    res = sorted(res.values(), key=lambda item: item[sort_key])
    return res


class FakeRequest(object):
    environ = {"nova.context": context.get_admin_context()}
    api_version_request = api_version.APIVersionRequest("2.1")

    def is_legacy_v2(self):
        return False


class FakeResponse(object):
    obj = {'flavor': {'id': '0'},
           'flavors': [
               {'id': '0'},
               {'id': '2'}]
    }

    def attach(self, **kwargs):
        pass


def fake_get_flavor_projects_from_db(context, flavorid):
    raise exception.FlavorNotFound(flavor_id=flavorid)


class FlavorAccessTestV21(test.NoDBTestCase):
    api_version = "2.1"
    FlavorAccessController = flavor_access_v21.FlavorAccessController
    FlavorActionController = flavor_access_v21.FlavorActionController
    _prefix = "/v2/%s" % fakes.FAKE_PROJECT_ID
    validation_ex = exception.ValidationError

    def setUp(self):
        super(FlavorAccessTestV21, self).setUp()
        self.flavor_controller = flavors_api.FlavorsController()
        # We need to stub out verify_project_id so that it doesn't
        # generate an EndpointNotFound exception and result in a
        # server error.
        self.stub_out('nova.api.openstack.identity.verify_project_id',
                      lambda ctx, project_id: True)

        self.req = FakeRequest()
        self.req.environ = {"nova.context": context.RequestContext(
            'fake_user', fakes.FAKE_PROJECT_ID)}
        self.stub_out('nova.objects.Flavor._flavor_get_by_flavor_id_from_db',
                      fake_get_flavor_by_flavor_id)
        self.stub_out('nova.objects.flavor._flavor_get_all_from_db',
                      fake_get_all_flavors_sorted_list)
        self.stub_out('nova.objects.flavor._get_projects_from_db',
                      fake_get_flavor_access_by_flavor_id)

        self.flavor_access_controller = self.FlavorAccessController()
        self.flavor_action_controller = self.FlavorActionController()

    def _verify_flavor_list(self, result, expected):
        # result already sorted by flavor_id
        self.assertEqual(len(result), len(expected))

        for d1, d2 in zip(result, expected):
            self.assertEqual(d1['id'], d2['id'])

    @mock.patch('nova.objects.Flavor._flavor_get_by_flavor_id_from_db',
                side_effect=exception.FlavorNotFound(flavor_id='foo'))
    def test_list_flavor_access_public(self, mock_api_get):
        # query os-flavor-access on public flavor should return 404
        self.assertRaises(exc.HTTPNotFound,
                          self.flavor_access_controller.index,
                          self.req, '1')

    def test_list_flavor_access_private(self):
        expected = {'flavor_access': [
            {'flavor_id': '2', 'tenant_id': 'proj2'},
            {'flavor_id': '2', 'tenant_id': 'proj3'}]}
        result = self.flavor_access_controller.index(self.req, '2')
        self.assertEqual(result, expected)

    def test_list_flavor_with_admin_default_proj1(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}]}
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors',
                                      use_admin_context=True)
        req.environ['nova.context'].project_id = 'proj1'
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_admin_default_proj2(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}, {'id': '2'}]}
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors',
                                      use_admin_context=True)
        req.environ['nova.context'].project_id = 'proj2'
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_admin_ispublic_true(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}]}
        url = self._prefix + '/flavors?is_public=true'
        req = fakes.HTTPRequest.blank(url,
                                      use_admin_context=True)
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_admin_ispublic_false(self):
        expected = {'flavors': [{'id': '2'}, {'id': '3'}]}
        url = self._prefix + '/flavors?is_public=false'
        req = fakes.HTTPRequest.blank(url,
                                      use_admin_context=True)
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_admin_ispublic_false_proj2(self):
        expected = {'flavors': [{'id': '2'}, {'id': '3'}]}
        url = self._prefix + '/flavors?is_public=false'
        req = fakes.HTTPRequest.blank(url,
                                      use_admin_context=True)
        req.environ['nova.context'].project_id = 'proj2'
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_admin_ispublic_none(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}, {'id': '2'},
                                {'id': '3'}]}
        url = self._prefix + '/flavors?is_public=none'
        req = fakes.HTTPRequest.blank(url,
                                      use_admin_context=True)
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_no_admin_default(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}]}
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors',
                                      use_admin_context=False)
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_no_admin_ispublic_true(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}]}
        url = self._prefix + '/flavors?is_public=true'
        req = fakes.HTTPRequest.blank(url,
                                      use_admin_context=False)
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_no_admin_ispublic_false(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}]}
        url = self._prefix + '/flavors?is_public=false'
        req = fakes.HTTPRequest.blank(url,
                                      use_admin_context=False)
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_list_flavor_with_no_admin_ispublic_none(self):
        expected = {'flavors': [{'id': '0'}, {'id': '1'}]}
        url = self._prefix + '/flavors?is_public=none'
        req = fakes.HTTPRequest.blank(url,
                                      use_admin_context=False)
        result = self.flavor_controller.index(req)
        self._verify_flavor_list(result['flavors'], expected['flavors'])

    def test_add_tenant_access(self):
        def stub_add_flavor_access(context, flavor_id, projectid):
            self.assertEqual(3, flavor_id, "flavor_id")
            self.assertEqual("proj2", projectid, "projectid")
        self.stub_out('nova.objects.Flavor._flavor_add_project',
                      stub_add_flavor_access)
        expected = {'flavor_access':
            [{'flavor_id': '3', 'tenant_id': 'proj3'}]}
        body = {'addTenantAccess': {'tenant': 'proj2'}}
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)

        result = self.flavor_action_controller._add_tenant_access(
            req, '3', body=body)
        self.assertEqual(result, expected)

    @mock.patch('nova.objects.Flavor.get_by_flavor_id',
                side_effect=exception.FlavorNotFound(flavor_id='1'))
    def test_add_tenant_access_with_flavor_not_found(self, mock_get):
        body = {'addTenantAccess': {'tenant': 'proj2'}}
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)
        self.assertRaises(exc.HTTPNotFound,
                          self.flavor_action_controller._add_tenant_access,
                          req, '2', body=body)

    def test_add_tenant_access_with_no_tenant(self):
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)
        body = {'addTenantAccess': {'foo': 'proj2'}}
        self.assertRaises(self.validation_ex,
                          self.flavor_action_controller._add_tenant_access,
                          req, '2', body=body)
        body = {'addTenantAccess': {'tenant': ''}}
        self.assertRaises(self.validation_ex,
                          self.flavor_action_controller._add_tenant_access,
                          req, '2', body=body)

    def test_add_tenant_access_with_already_added_access(self):
        def stub_add_flavor_access(context, flavorid, projectid):
            raise exception.FlavorAccessExists(flavor_id=flavorid,
                                               project_id=projectid)
        self.stub_out('nova.objects.Flavor._flavor_add_project',
                      stub_add_flavor_access)
        body = {'addTenantAccess': {'tenant': 'proj2'}}
        self.assertRaises(exc.HTTPConflict,
                          self.flavor_action_controller._add_tenant_access,
                          self.req, '3', body=body)

    def test_remove_tenant_access_with_bad_access(self):
        def stub_remove_flavor_access(context, flavorid, projectid):
            raise exception.FlavorAccessNotFound(flavor_id=flavorid,
                                                 project_id=projectid)
        self.stub_out('nova.objects.Flavor._flavor_del_project',
                      stub_remove_flavor_access)
        body = {'removeTenantAccess': {'tenant': 'proj2'}}
        self.assertRaises(exc.HTTPNotFound,
                          self.flavor_action_controller._remove_tenant_access,
                          self.req, '3', body=body)

    def test_add_tenant_access_is_public(self):
        body = {'addTenantAccess': {'tenant': 'proj2'}}
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)
        req.api_version_request = api_version.APIVersionRequest('2.7')
        self.assertRaises(exc.HTTPConflict,
                          self.flavor_action_controller._add_tenant_access,
                          req, '1', body=body)

    @mock.patch('nova.objects.Flavor._flavor_get_by_flavor_id_from_db',
                side_effect=exception.FlavorNotFound(flavor_id='foo'))
    def test_delete_tenant_access_with_no_tenant(self, mock_api_get):
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)
        body = {'removeTenantAccess': {'foo': 'proj2'}}
        self.assertRaises(self.validation_ex,
                          self.flavor_action_controller._remove_tenant_access,
                          req, '2', body=body)
        body = {'removeTenantAccess': {'tenant': ''}}
        self.assertRaises(self.validation_ex,
                          self.flavor_action_controller._remove_tenant_access,
                          req, '2', body=body)

    @mock.patch('nova.api.openstack.identity.verify_project_id',
                side_effect=exc.HTTPBadRequest(
                    explanation="Project ID proj2 is not a valid project."))
    def test_add_tenant_access_with_invalid_tenant(self, mock_verify):
        """Tests the case that the tenant does not exist in Keystone."""
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)
        body = {'addTenantAccess': {'tenant': 'proj2'}}
        self.assertRaises(exc.HTTPBadRequest,
                          self.flavor_action_controller._add_tenant_access,
                          req, '2', body=body)
        mock_verify.assert_called_once_with(
            req.environ['nova.context'], 'proj2')

    @mock.patch('nova.objects.Flavor.remove_access')
    @mock.patch('nova.api.openstack.identity.verify_project_id',
                side_effect=exc.HTTPBadRequest(
                    explanation="Project ID proj2 is not a valid project."))
    def test_remove_tenant_access_with_invalid_tenant(self,
                                                      mock_verify,
                                                      mock_remove_access):
        """Tests the case that the tenant does not exist in Keystone."""
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)
        body = {'removeTenantAccess': {'tenant': 'proj2'}}

        self.flavor_action_controller._remove_tenant_access(
            req, '2', body=body)
        mock_verify.assert_called_once_with(
            req.environ['nova.context'], 'proj2')
        mock_remove_access.assert_called_once_with('proj2')

    @mock.patch('nova.api.openstack.identity.verify_project_id',
                side_effect=exc.HTTPBadRequest(
                    explanation="Nova was unable to find Keystone "
                                "service endpoint."))
    def test_remove_tenant_access_missing_keystone_endpoint(self,
                                                            mock_verify):
        """Tests the case that Keystone identity service endpoint
        version 3.0 was not found.
        """
        req = fakes.HTTPRequest.blank(self._prefix + '/flavors/2/action',
                                      use_admin_context=True)
        body = {'removeTenantAccess': {'tenant': 'proj2'}}

        self.assertRaises(exc.HTTPBadRequest,
                          self.flavor_action_controller._remove_tenant_access,
                          req, '2', body=body)
        mock_verify.assert_called_once_with(
            req.environ['nova.context'], 'proj2')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import glob
import jsonschema
import logging
import microversion_parse
import os
import yaml

import os_resource_classes
import os_traits

from nova import exception as nova_exc
from nova.i18n import _

LOG = logging.getLogger(__name__)

# A dictionary with keys for all supported major versions with lists of
# corresponding minor versions as values.
SUPPORTED_SCHEMA_VERSIONS = {
    1: {0}
}

# Supported provider config file schema
SCHEMA_V1 = {
    # This defintion uses JSON Schema Draft 7.
    # https://json-schema.org/draft-07/json-schema-release-notes.html
    'type': 'object',
    'properties': {
        # This property is used to track where the provider.yaml file
        # originated. It is reserved for internal use and should never be
        # set in a provider.yaml file supplied by an end user.
        '__source_file': {'not': {}},
        'meta': {
            'type': 'object',
            'properties': {
                # Version ($Major, $minor) of the schema must successfully
                # parse documents conforming to ($Major, 0..N).
                # Any breaking schema change (e.g. removing fields, adding
                # new required fields, imposing a stricter pattern on a value,
                # etc.) must bump $Major.
                'schema_version': {
                    'type': 'string',
                    'pattern': '^1.([0-9]|[1-9][0-9]+)$'
                }
            },
            'required': ['schema_version'],
            'additionalProperties': True
        },
        'providers': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'identification': {
                        '$ref': '#/$defs/providerIdentification'
                    },
                    'inventories': {
                        '$ref': '#/$defs/providerInventories'
                    },
                    'traits': {
                        '$ref': '#/$defs/providerTraits'
                    }
                },
                'required': ['identification'],
                'additionalProperties': True
            }
        }
    },
    'required': ['meta'],
    'additionalProperties': True,
    '$defs': {
        'providerIdentification': {
            # Identify a single provider to configure.
            # Exactly one identification method should be used. Currently
            # `uuid` or `name` are supported, but future versions may
            # support others. The uuid can be set to the sentinel value
            # `$COMPUTE_NODE` which will cause the consuming compute service to
            # apply the configuration to all compute node root providers
            # it manages that are not otherwise specified using a uuid or name.
            'type': 'object',
            'properties': {
                'uuid': {
                    'oneOf': [
                        {
                            # TODO(sean-k-mooney): replace this with type uuid
                            # when we can depend on a version of the jsonschema
                            # lib that implements draft 8 or later of the
                            # jsonschema spec.
                            'type': 'string',
                            'pattern':
                                '^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-'
                                '[0-9A-Fa-f]{4}-[0-9A-Fa-f]{4}-'
                                '[0-9A-Fa-f]{12}$'
                        },
                        {
                            'type': 'string',
                            'const': '$COMPUTE_NODE'
                        }
                    ]
                },
                'name': {
                    'type': 'string',
                    'minLength': 1,
                    'maxLength': 200
                }
            },
            # This introduces the possibility of an unsupported key name being
            # used to get by schema validation, but is necessary to support
            # forward compatibility with new identification methods.
            # This should be checked after schema validation.
            'minProperties': 1,
            'maxProperties': 1,
            'additionalProperties': False
        },
        'providerInventories': {
            # Allows the admin to specify various adjectives to create and
            # manage providers' inventories. This list of adjectives can be
            # extended in the future as the schema evolves to meet new use
            # cases. As of v1.0, only one adjective, `additional`, is
            # supported.
            'type': 'object',
            'properties': {
                'additional': {
                    'type': 'array',
                    'items': {
                        'patternProperties': {
                            # Allows any key name matching the resource class
                            # pattern, check to prevent conflicts with virt
                            # driver owned resouces classes will be done after
                            # schema validation.
                            '^[A-Z0-9_]{1,255}$': {
                                'type': 'object',
                                'properties': {
                                    # Any optional properties not populated
                                    # will be given a default value by
                                    # placement. If overriding a pre-existing
                                    # provider values will not be preserved
                                    # from the existing inventory.
                                    'total': {
                                        'type': 'integer'
                                    },
                                    'reserved': {
                                        'type': 'integer'
                                    },
                                    'min_unit': {
                                        'type': 'integer'
                                    },
                                    'max_unit': {
                                        'type': 'integer'
                                    },
                                    'step_size': {
                                        'type': 'integer'
                                    },
                                    'allocation_ratio': {
                                        'type': 'number'
                                    }
                                },
                                'required': ['total'],
                                # The defined properties reflect the current
                                # placement data model. While defining those
                                # in the schema and not allowing additional
                                # properties means we will need to bump the
                                # schema version if they change, that is likely
                                # to be part of a large change that may have
                                # other impacts anyway. The benefit of stricter
                                # validation of property names outweighs the
                                # (small) chance of having to bump the schema
                                # version as described above.
                                'additionalProperties': False
                            }
                        },
                        # This ensures only keys matching the pattern
                        # above are allowed.
                        'additionalProperties': False
                    }
                }
            },
            'additionalProperties': True
        },
        'providerTraits': {
            # Allows the admin to specify various adjectives to create and
            # manage providers' traits. This list of adjectives can be extended
            # in the future as the schema evolves to meet new use cases.
            # As of v1.0, only one adjective, `additional`, is supported.
            'type': 'object',
            'properties': {
                'additional': {
                    'type': 'array',
                    'items': {
                        # Allows any value matching the trait pattern here,
                        # additional validation will be done after schema
                        # validation.
                        'type': 'string',
                        'pattern': '^[A-Z0-9_]{1,255}$'
                    }
                }
            },
            'additionalProperties': True
        }
    }
}


def _load_yaml_file(path):
    """Loads and parses a provider.yaml config file into a dict.

    :param path: Path to the yaml file to load.
    :return: Dict representing the yaml file requested.
    :raise: ProviderConfigException if the path provided cannot be read
            or the file is not valid yaml.
    """
    try:
        with open(path) as open_file:
            try:
                return yaml.safe_load(open_file)
            except yaml.YAMLError as ex:
                message = _("Unable to load yaml file: %s ") % ex
                if hasattr(ex, 'problem_mark'):
                    pos = ex.problem_mark
                    message += _("File: %s ") % open_file.name
                    message += _("Error position: (%s:%s)") % (
                        pos.line + 1, pos.column + 1)
                raise nova_exc.ProviderConfigException(error=message)
    except OSError:
        message = _("Unable to read yaml config file: %s") % path
        raise nova_exc.ProviderConfigException(error=message)


def _validate_provider_config(config, provider_config_path):
    """Accepts a schema-verified provider config in the form of a dict and
     performs additional checks for format and required keys.

    :param config: Dict containing a provider config file
    :param provider_config_path: Path to the provider config, used for logging
    :return: List of valid providers
    :raise nova.exception.ProviderConfigException: If provider id is missing,
        or a resource class or trait name is invalid.
    """
    def _validate_traits(provider):
        # Check that traits are custom
        additional_traits = set(provider.get("traits", {}).get(
            "additional", []))
        trait_conflicts = [trait for trait in additional_traits
                           if not os_traits.is_custom(trait)]
        if trait_conflicts:
            # sort for more predictable message for testing
            message = _(
                "Invalid traits, only custom traits are allowed: %s"
            ) % sorted(trait_conflicts)
            raise nova_exc.ProviderConfigException(error=message)
        return additional_traits

    def _validate_rc(provider):
        # Check that resource classes are custom
        additional_inventories = provider.get("inventories", {}).get(
            "additional", [])
        all_inventory_conflicts = []
        for inventory in additional_inventories:
            inventory_conflicts = [rc for rc in inventory
                                   if not os_resource_classes.is_custom(rc)]
            all_inventory_conflicts += inventory_conflicts

        if all_inventory_conflicts:
            # sort for more predictable message for testing
            message = _(
                "Invalid resource class, only custom resource classes "
                "are allowed: %s") % ', '.join(sorted(all_inventory_conflicts))
            raise nova_exc.ProviderConfigException(error=message)
        return additional_inventories

    # store valid providers
    valid_providers = []

    for provider in config.get("providers", []):
        # Check that the identification method is known since
        # the schema only requires that some property be present
        pid = provider["identification"]
        provider_id = pid.get("name") or pid.get("uuid")
        # Not checking the validity of provider_id since
        # the schema has already ensured that.

        additional_traits = _validate_traits(provider)
        additional_inventories = _validate_rc(provider)

        # filter out no-op providers so they will not be returned
        if not additional_traits and not additional_inventories:
            message = (
                "Provider %(provider_id)s defined in %(provider_config_path)s "
                "has no additional inventories or traits and will be ignored."
            ) % {
                "provider_id": provider_id,
                "provider_config_path": provider_config_path
            }
            LOG.warning(message)
        else:
            valid_providers.append(provider)

    return valid_providers


def _parse_provider_yaml(path):
    """Loads schema, parses a provider.yaml file and validates the content.

    :param path: File system path to the file to parse.
    :return: dict representing the contents of the file.
    :raise ProviderConfigException: If the specified file does
        not validate against the schema, the schema version is not supported,
        or if unable to read configuration or schema files.
    """
    yaml_file = _load_yaml_file(path)

    try:
        schema_version = microversion_parse.parse_version_string(
            yaml_file['meta']['schema_version'])
    except (KeyError, TypeError):
        message = _("Unable to detect schema version: %s") % yaml_file
        raise nova_exc.ProviderConfigException(error=message)

    if schema_version.major not in SUPPORTED_SCHEMA_VERSIONS:
        message = _(
            "Unsupported schema major version: %d") % schema_version.major
        raise nova_exc.ProviderConfigException(error=message)

    if schema_version.minor not in \
            SUPPORTED_SCHEMA_VERSIONS[schema_version.major]:
        # TODO(sean-k-mooney): We should try to provide a better
        # message that identifies which fields may be ignored
        # and the max minor version supported by this version of nova.
        message = (
            "Provider config file [%(path)s] is at schema version "
            "%(schema_version)s. Nova supports the major version, "
            "but not the minor. Some fields may be ignored."
            % {"path": path, "schema_version": schema_version})
        LOG.warning(message)

    try:
        jsonschema.validate(yaml_file, SCHEMA_V1)
    except jsonschema.exceptions.ValidationError as e:
        message = _(
            "The provider config file %(path)s did not pass validation "
            "for schema version %(schema_version)s: %(reason)s") % {
                "path": path, "schema_version": schema_version, "reason": e}
        raise nova_exc.ProviderConfigException(error=message)
    return yaml_file


def get_provider_configs(provider_config_dir):
    """Gathers files in the provided path and calls the parser for each file
    and merges them into a list while checking for a number of possible
    conflicts.

    :param provider_config_dir: Path to a directory containing provider config
        files to be loaded.
    :raise nova.exception.ProviderConfigException: If unable to read provider
        config directory or if one of a number of validation checks fail:
        - Unknown, unsupported, or missing schema major version.
        - Unknown, unsupported, or missing resource provider identification.
        - A specific resource provider is identified twice with the same
          method. If the same provider identified by *different* methods,
          such conflict will be detected in a later stage.
        - A resource class or trait name is invalid or not custom.
        - A general schema validation error occurs (required fields,
          types, etc).
    :return: A dict of dicts keyed by uuid_or_name with the parsed and
    validated contents of all files in the provided dir. Each value in the dict
    will include the source file name the value of the __source_file key.
    """
    provider_configs = {}

    provider_config_paths = glob.glob(
        os.path.join(provider_config_dir, "*.yaml"))
    provider_config_paths.sort()

    if not provider_config_paths:
        message = (
            "No provider configs found in %s. If files are present, "
            "ensure the Nova process has access."
        )
        LOG.info(message, provider_config_dir)

        # return an empty dict as no provider configs found
        return provider_configs

    for provider_config_path in provider_config_paths:
        provider_config = _parse_provider_yaml(provider_config_path)

        for provider in _validate_provider_config(
            provider_config, provider_config_path,
        ):
            provider['__source_file'] = os.path.basename(provider_config_path)
            pid = provider["identification"]
            uuid_or_name = pid.get("uuid") or pid.get("name")
            # raise exception if this provider was already processed
            if uuid_or_name in provider_configs:
                raise nova_exc.ProviderConfigException(
                    error=_(
                        "Provider %(provider_id)s has multiple definitions "
                        "in source file(s): %(source_files)s."
                    ) % {
                        "provider_id": uuid_or_name,
                        # sorted set for deduplication and consistent order
                        "source_files": sorted({
                            provider_configs[uuid_or_name]["__source_file"],
                            provider_config_path
                        })
                    }
                )
            provider_configs[uuid_or_name] = provider
    return provider_configs



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_serialization import jsonutils

from nova.api.openstack import api_version_request as api_version
from nova import test
from nova.tests.unit.api.openstack.compute import microversions
from nova.tests.unit.api.openstack import fakes


class LegacyMicroversionsTest(test.NoDBTestCase):

    header_name = 'X-OpenStack-Nova-API-Version'

    def setUp(self):
        super(LegacyMicroversionsTest, self).setUp()
        self.app = fakes.wsgi_app_v21(custom_routes=microversions.ROUTES)

    def _test_microversions(self, app, req, ret_code, ret_header=None):
        req.environ['CONTENT_TYPE'] = "application/json"

        res = req.get_response(app)
        self.assertEqual(ret_code, res.status_int)
        if ret_header:
            if 'nova' not in self.header_name.lower():
                ret_header = 'compute %s' % ret_header
            self.assertEqual(ret_header,
                             res.headers[self.header_name])
        return res

    def _make_header(self, req_header):
        if 'nova' in self.header_name.lower():
            headers = {self.header_name: req_header}
        else:
            headers = {self.header_name: 'compute %s' % req_header}
        return headers

    def test_microversions_no_header(self):
        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions' % fakes.FAKE_PROJECT_ID,
                method='GET')
        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual('val', resp_json['param'])

    def test_microversions_return_header(self):
        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions' % fakes.FAKE_PROJECT_ID)
        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual('val', resp_json['param'])
        if 'nova' in self.header_name.lower():
            self.assertEqual("2.1", res.headers[self.header_name])
        else:
            self.assertEqual("compute 2.1", res.headers[self.header_name])
        self.assertIn(self.header_name, res.headers.getall('Vary'))

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions_return_header_non_default(self,
                                                     mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("2.3")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions' % fakes.FAKE_PROJECT_ID)
        req.headers = self._make_header('2.3')
        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual('val2', resp_json['param'])
        if 'nova' in self.header_name.lower():
            self.assertEqual("2.3", res.headers[self.header_name])
        else:
            self.assertEqual("compute 2.3", res.headers[self.header_name])
        self.assertIn(self.header_name, res.headers.getall('Vary'))

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions_return_header_fault(self, mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.0")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions' % fakes.FAKE_PROJECT_ID)
        req.headers = self._make_header('3.0')
        res = req.get_response(self.app)
        self.assertEqual(400, res.status_int)
        if 'nova' in self.header_name.lower():
            self.assertEqual("3.0", res.headers[self.header_name])
        else:
            self.assertEqual("compute 3.0", res.headers[self.header_name])
        self.assertIn(self.header_name, res.headers.getall('Vary'))

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def _check_microversion_response(self, url, req_version, resp_param,
                                     mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest('2.3')

        req = fakes.HTTPRequest.blank(url)
        req.headers = self._make_header(req_version)
        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual(resp_param, resp_json['param'])

    def test_microversions_with_header(self):
        self._check_microversion_response(
            '/v2/%s/microversions' % fakes.FAKE_PROJECT_ID,
            '2.3', 'val2')

    def test_microversions_with_header_exact_match(self):
        self._check_microversion_response(
            '/v2/%s/microversions' % fakes.FAKE_PROJECT_ID,
            '2.2', 'val2')

    def test_microversions2_no_2_1_version(self):
        self._check_microversion_response(
            '/v2/%s/microversions2' % fakes.FAKE_PROJECT_ID,
            '2.3', 'controller2_val1')

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions2_later_version(self, mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.1")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions2' % fakes.FAKE_PROJECT_ID)
        req.headers = self._make_header('3.0')
        res = req.get_response(self.app)
        self.assertEqual(202, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual('controller2_val2', resp_json['param'])

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions2_version_too_high(self, mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.5")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions2' % fakes.FAKE_PROJECT_ID)
        req.headers = {self.header_name: '3.2'}
        res = req.get_response(self.app)
        self.assertEqual(404, res.status_int)

    def test_microversions2_version_too_low(self):
        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions2' % fakes.FAKE_PROJECT_ID)
        req.headers = {self.header_name: '2.1'}
        res = req.get_response(self.app)
        self.assertEqual(404, res.status_int)

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions_global_version_too_high(self,
                                                   mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.5")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions2' % fakes.FAKE_PROJECT_ID)
        req.headers = self._make_header('3.7')
        res = req.get_response(self.app)
        self.assertEqual(406, res.status_int)
        res_json = jsonutils.loads(res.body)
        self.assertEqual("Version 3.7 is not supported by the API. "
                         "Minimum is 2.1 and maximum is 3.5.",
                         res_json['computeFault']['message'])

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions_schema(self, mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.3")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions3' % fakes.FAKE_PROJECT_ID)
        req.method = 'POST'
        req.headers = self._make_header('2.2')
        req.environ['CONTENT_TYPE'] = "application/json"
        req.body = jsonutils.dump_as_bytes({'dummy': {'val': 'foo'}})

        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual('create_val1', resp_json['param'])
        if 'nova' in self.header_name.lower():
            self.assertEqual("2.2", res.headers[self.header_name])
        else:
            self.assertEqual("compute 2.2", res.headers[self.header_name])
        self.assertIn(self.header_name, res.headers.getall('Vary'))

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions_schema_fail(self, mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.3")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions3' % fakes.FAKE_PROJECT_ID)
        req.method = 'POST'
        req.headers = {self.header_name: '2.2'}
        req.environ['CONTENT_TYPE'] = "application/json"
        req.body = jsonutils.dump_as_bytes({'dummy': {'invalid_param': 'foo'}})

        res = req.get_response(self.app)
        self.assertEqual(400, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertTrue(resp_json['badRequest']['message'].startswith(
            "Invalid input for field/attribute dummy."))

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions_schema_out_of_version_check(self,
                                                       mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.3")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions3/1' % fakes.FAKE_PROJECT_ID)
        req.method = 'PUT'
        req.headers = self._make_header('2.2')
        req.body = jsonutils.dump_as_bytes({'dummy': {'inv_val': 'foo'}})
        req.environ['CONTENT_TYPE'] = "application/json"

        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual('update_val1', resp_json['param'])
        if 'nova' in self.header_name.lower():
            self.assertEqual("2.2", res.headers[self.header_name])
        else:
            self.assertEqual("compute 2.2", res.headers[self.header_name])

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def test_microversions_schema_second_version(self,
                                                 mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("3.3")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions3/1' % fakes.FAKE_PROJECT_ID)
        req.headers = self._make_header('2.10')
        req.environ['CONTENT_TYPE'] = "application/json"
        req.method = 'PUT'
        req.body = jsonutils.dump_as_bytes({'dummy': {'val2': 'foo'}})

        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual('update_val1', resp_json['param'])
        if 'nova' in self.header_name.lower():
            self.assertEqual("2.10", res.headers[self.header_name])
        else:
            self.assertEqual("compute 2.10", res.headers[self.header_name])

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def _test_microversions_inner_function(self, version, expected_resp,
                                           mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("2.2")
        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions4' % fakes.FAKE_PROJECT_ID)
        req.headers = self._make_header(version)
        req.environ['CONTENT_TYPE'] = "application/json"
        req.method = 'POST'
        req.body = b''

        res = req.get_response(self.app)
        self.assertEqual(200, res.status_int)
        resp_json = jsonutils.loads(res.body)
        self.assertEqual(expected_resp, resp_json['param'])
        if 'nova' not in self.header_name.lower():
            version = 'compute %s' % version
        self.assertEqual(version, res.headers[self.header_name])

    def test_microversions_inner_function_v22(self):
        self._test_microversions_inner_function('2.2', 'controller4_val2')

    def test_microversions_inner_function_v21(self):
        self._test_microversions_inner_function('2.1', 'controller4_val1')

    @mock.patch("nova.api.openstack.api_version_request.max_api_version")
    def _test_microversions_actions(self, ret_code, ret_header, req_header,
                                    mock_maxver):
        mock_maxver.return_value = api_version.APIVersionRequest("2.3")

        req = fakes.HTTPRequest.blank(
                '/v2/%s/microversions3/1/action' % fakes.FAKE_PROJECT_ID)
        if req_header:
            req.headers = self._make_header(req_header)
        req.method = 'POST'
        req.body = jsonutils.dump_as_bytes({'foo': None})

        res = self._test_microversions(self.app, req, ret_code,
                                       ret_header=ret_header)
        if ret_code == 202:
            resp_json = jsonutils.loads(res.body)
            self.assertEqual({'foo': 'bar'}, resp_json)

    def test_microversions_actions(self):
        self._test_microversions_actions(202, "2.1", "2.1")

    def test_microversions_actions_too_high(self):
        self._test_microversions_actions(404, "2.3", "2.3")

    def test_microversions_actions_no_header(self):
        self._test_microversions_actions(202, "2.1", None)


class MicroversionsTest(LegacyMicroversionsTest):

    header_name = 'OpenStack-API-Version'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from nova.tests.functional import integrated_helpers


class TestServices(integrated_helpers._IntegratedTestBase):
    api_major_version = 'v2.1'
    microversion = 'latest'

    def setUp(self):
        super(TestServices, self).setUp()
        self.compute_rp_uuid = self.admin_api.api_get(
            'os-hypervisors?hypervisor_hostname_pattern=fake-mini'
        ).body['hypervisors'][0]['id']
        self.compute_service_id = self.admin_api.get_services(
            host='compute', binary='nova-compute')[0]['id']

    def _get_traits_on_compute(self):
        return self.placement.get(
            '/resource_providers/%s/traits' % self.compute_rp_uuid,
            version='1.6'
        ).body['traits']

    def _disable_compute(self):
        self.admin_api.put_service(
            self.compute_service_id, {'status': 'disabled'})

    def _enable_compute(self):
        self.admin_api.put_service(
            self.compute_service_id, {'status': 'enabled'})

    def _has_disabled_trait(self):
        return "COMPUTE_STATUS_DISABLED" in self._get_traits_on_compute()

    def test_compute_disable_after_server_create(self):
        # Check that COMPUTE_STATUS_DISABLED is not on the compute
        self.assertFalse(self._has_disabled_trait())

        self._disable_compute()
        # Check that COMPUTE_STATUS_DISABLED is now on the compute
        self.assertTrue(self._has_disabled_trait())

        self._enable_compute()
        # Check that COMPUTE_STATUS_DISABLED is not on the compute
        self.assertFalse(self._has_disabled_trait())

        # Create a server.
        self._create_server(networks=[])

        self._disable_compute()

        # Check that COMPUTE_STATUS_DISABLED is now on the compute.
        self.assertTrue(self._has_disabled_trait())

        # This would be the expected behavior
        #
        # self.assertTrue(self._has_disabled_trait())
        #
        # Alternatively the test could wait for the periodic to run or trigger
        # it manually.

        # This passes now but not because enabling works but because the
        # above fault caused that COMPUTE_STATUS_DISABLED is not on the compute
        # RP in the first place.
        self._enable_compute()
        # Check that COMPUTE_STATUS_DISABLED is removed from the compute
        self.assertFalse(self._has_disabled_trait())
        self.assertNotIn(
            'An error occurred while updating COMPUTE_STATUS_DISABLED trait '
            'on compute node resource provider',
            self.stdlog.logger.output,
            "This is probably bug 1886418.")



# Copyright 2014, 2017 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from oslo_utils.fixture import uuidsentinel

from nova.compute import power_state
from nova.compute import vm_states
from nova import objects

try:
    import powervm  # noqa: F401
except ImportError:
    raise unittest.SkipTest(
        "The 'pypowervm' dependency is not installed."
    )


TEST_FLAVOR = objects.flavor.Flavor(
    memory_mb=2048,
    swap=0,
    vcpu_weight=None,
    root_gb=10, id=2,
    name=u'm1.small',
    ephemeral_gb=0,
    rxtx_factor=1.0,
    flavorid=uuidsentinel.flav_id,
    vcpus=1)

TEST_INSTANCE = objects.Instance(
    id=1,
    uuid=uuidsentinel.inst_id,
    display_name='Fake Instance',
    root_gb=10,
    ephemeral_gb=0,
    instance_type_id=TEST_FLAVOR.id,
    system_metadata={'image_os_distro': 'rhel'},
    host='host1',
    flavor=TEST_FLAVOR,
    task_state=None,
    vm_state=vm_states.STOPPED,
    power_state=power_state.SHUTDOWN,
)

IMAGE1 = {
    'id': uuidsentinel.img_id,
    'name': 'image1',
    'size': 300,
    'container_format': 'bare',
    'disk_format': 'raw',
    'checksum': 'b518a8ba2b152b5607aceb5703fac072',
}
TEST_IMAGE1 = objects.image_meta.ImageMeta.from_dict(IMAGE1)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova.api.openstack.compute import hosts
from nova.policies import base as base_policy
from nova.policies import hosts as policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.policies import base


class HostsPolicyTest(base.BasePolicyTest):
    """Test os-hosts APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(HostsPolicyTest, self).setUp()
        self.controller = hosts.HostController()
        self.req = fakes.HTTPRequest.blank('')

        # With legacy rule and scope check disabled by default, system admin,
        # legacy admin, and project admin will be able to perform hosts
        # Operations.
        self.system_admin_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context]

    @mock.patch('nova.compute.api.HostAPI.service_get_all')
    def test_list_hosts_policy(self, mock_get):
        rule_name = policies.POLICY_NAME % 'list'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.index,
                                self.req)

    @mock.patch('nova.context.set_target_cell')
    @mock.patch('nova.objects.HostMapping.get_by_host')
    @mock.patch('nova.objects.ComputeNode.'
        'get_first_node_by_host_for_old_compat')
    @mock.patch('nova.compute.api.HostAPI.instance_get_all_by_host')
    def test_show_host_policy(self, mock_get, mock_node, mock_map, mock_set):
        rule_name = policies.POLICY_NAME % 'show'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.show,
                                self.req, 11111)

    def test_update_host_policy(self):
        rule_name = policies.POLICY_NAME % 'update'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.update,
                                self.req, 11111, body={})

    @mock.patch('nova.compute.api.HostAPI.host_power_action')
    def test_reboot_host_policy(self, mock_action):
        rule_name = policies.POLICY_NAME % 'reboot'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.reboot,
                                self.req, 11111)

    @mock.patch('nova.compute.api.HostAPI.host_power_action')
    def test_shutdown_host_policy(self, mock_action):
        rule_name = policies.POLICY_NAME % 'shutdown'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.shutdown,
                                self.req, 11111)

    @mock.patch('nova.compute.api.HostAPI.host_power_action')
    def test_startup_host_policy(self, mock_action):
        rule_name = policies.POLICY_NAME % 'start'
        self.common_policy_auth(self.system_admin_authorized_contexts,
                                rule_name, self.controller.startup,
                                self.req, 11111)


class HostsNoLegacyNoScopePolicyTest(HostsPolicyTest):
    """Test Hosts APIs policies with no legacy deprecated rules
    and no scope checks which means new defaults only. In this case
    system admin, legacy admin, and project admin will be able to perform
    hosts Operations. Legacy admin will be allowed as policy is just admin
    if no scope checks.

    """

    without_deprecated_rules = True


class HostsScopeTypePolicyTest(HostsPolicyTest):
    """Test os-hosts APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(HostsScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # With scope checks enable, only system admin is able to perform
        # hosts Operations.
        self.system_admin_authorized_contexts = [self.system_admin_context]


class HostsScopeTypeNoLegacyPolicyTest(HostsScopeTypePolicyTest):
    """Test Hosts APIs policies with with no legacy deprecated rules
    and scope checks enabled which means scope + new defaults. So
    only system admin is able to perform hosts Operations.
    """

    without_deprecated_rules = True
    rules_without_deprecation = {
        policies.POLICY_NAME % 'list':
            base_policy.ADMIN,
        policies.POLICY_NAME % 'show':
            base_policy.ADMIN,
        policies.POLICY_NAME % 'update':
            base_policy.ADMIN,
        policies.POLICY_NAME % 'reboot':
            base_policy.ADMIN,
        policies.POLICY_NAME % 'shutdown':
            base_policy.ADMIN,
        policies.POLICY_NAME % 'startup':
            base_policy.ADMIN}



# Copyright (c) 2014-2016 Red Hat, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class SecurityProxy(metaclass=abc.ABCMeta):
    """A console security Proxy Helper

    Console security proxy helpers should subclass
    this class and implement a generic `connect`
    for the particular protocol being used.

    Security drivers can then subclass the
    protocol-specific helper class.
    """

    @abc.abstractmethod
    def connect(self, tenant_sock, compute_sock):
        """Initiate the console connection

        This method performs the protocol specific
        negotiation, and returns the socket-like
        object to use to communicate with the server
        securely.

        :param tenant_sock: socket connected to the remote tenant user
        :param compute_sock: socket connected to the compute node instance

        :returns: a new compute_sock for the instance
        """
        pass



# Copyright (c) 2015 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova.tests.unit.virt.libvirt.volume import test_volume
from nova.virt.libvirt.volume import scaleio


class LibvirtScaleIOVolumeDriverTestCase(
        test_volume.LibvirtVolumeBaseTestCase):

    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory',
        new=mock.Mock(return_value=mock.Mock()))
    def test_libvirt_scaleio_driver_connect(self):
        sio = scaleio.LibvirtScaleIOVolumeDriver(self.fake_host)
        disk_info = {'path': '/dev/vol01', 'name': 'vol01'}
        conn = {'data': disk_info}
        sio.connector.connect_volume.return_value = conn['data']
        sio.connect_volume(conn, mock.sentinel.instance)
        sio.connector.connect_volume.assert_called_once_with(conn['data'])
        self.assertEqual(disk_info['path'], conn['data']['device_path'])

    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory',
        new=mock.Mock(return_value=mock.Mock()))
    def test_libvirt_scaleio_driver_get_config(self):
        sio = scaleio.LibvirtScaleIOVolumeDriver(self.fake_host)
        conn = {'data': {'device_path': '/dev/vol01'}}
        conf = sio.get_config(conn, self.disk_info)
        self.assertEqual('block', conf.source_type)
        self.assertEqual('/dev/vol01', conf.source_path)

    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory',
        new=mock.Mock(return_value=mock.Mock()))
    def test_libvirt_scaleio_driver_disconnect(self):
        sio = scaleio.LibvirtScaleIOVolumeDriver(self.fake_host)
        conn = {'data': mock.sentinel.conn_data}
        sio.disconnect_volume(conn, mock.sentinel.instance)
        sio.connector.disconnect_volume.assert_called_once_with(
            mock.sentinel.conn_data, None, force=False)

        # Verify force=True
        sio.connector.disconnect_volume.reset_mock()
        sio.disconnect_volume(conn, mock.sentinel.instance, force=True)
        sio.connector.disconnect_volume.assert_called_once_with(
            mock.sentinel.conn_data, None, force=True)

    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory',
        new=mock.Mock(return_value=mock.Mock()))
    def test_libvirt_scaleio_driver_extend_volume(self):
        extended_vol_size = 8
        sio = scaleio.LibvirtScaleIOVolumeDriver(self.fake_host)
        disk_info = {'size': extended_vol_size,
                     'name': 'vol01',
                     'device_path': '/dev/vol01'}
        conn = {'data': disk_info}
        sio.connector.extend_volume.return_value = extended_vol_size
        self.assertEqual(
            extended_vol_size,
            sio.extend_volume(conn, mock.sentinel.instance, extended_vol_size))



# Copyright (c) 2014 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Middleware that ensures x-compute-request-id

Nova's notion of request-id tracking predates any common idea, so the
original version of this header in OpenStack was
x-compute-request-id. Eventually we got oslo, and all other projects
implemented this with x-openstack-request-id.

However, x-compute-request-id was always part of our contract. The
following migrates us to use x-openstack-request-id as well, by using
the common middleware.

"""

from oslo_middleware import request_id


HTTP_RESP_HEADER_REQUEST_ID = 'x-compute-request-id'


class ComputeReqIdMiddleware(request_id.RequestId):
    compat_headers = [HTTP_RESP_HEADER_REQUEST_ID]
